"""Builder for CVs."""

import datetime as dt
from copy import copy, deepcopy

from dateutil import parser as date_parser
from dateutil.relativedelta import relativedelta

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.builders.cpbuilder import is_declined, is_pending
from regolith.dates import get_dates, is_current
from regolith.fsclient import _id_key
from regolith.sorters import doc_date_key, position_key
from regolith.stylers import month_fullnames, sentencecase
from regolith.tools import (
    all_docs_from_collection,
    awards,
    filter_activities,
    filter_committees,
    filter_employment_for_advisees,
    filter_facilities,
    filter_grants,
    filter_licenses,
    filter_patents,
    filter_presentations,
    filter_projects,
    filter_publications,
    filter_service,
    fuzzy_retrieval,
    get_id_from_name,
    make_bibtex_file,
    merge_collections_all,
)


class ActivitylogBuilder(LatexBuilderBase):
    """Build CV from database entries"""

    btype = "annual-activity"
    needed_colls = [
        "groups",
        "people",
        "grants",
        "proposals",
        "institutions",
        "projects",
        "presentations",
        "patents",
        "citations",
        "proposalReviews",
        "refereeReports",
        "recletts",
        "contacts",
    ]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["institutions"] = sorted(all_docs_from_collection(rc.client, "institutions"), key=_id_key)
        gtx["contacts"] = sorted(all_docs_from_collection(rc.client, "institutions"), key=_id_key)
        gtx["groups"] = sorted(all_docs_from_collection(rc.client, "groups"), key=_id_key)
        gtx["grants"] = sorted(all_docs_from_collection(rc.client, "grants"), key=_id_key)
        gtx["proposals"] = sorted(all_docs_from_collection(rc.client, "proposals"), key=_id_key)
        gtx["projects"] = sorted(all_docs_from_collection(rc.client, "projects"), key=_id_key)
        gtx["presentations"] = sorted(all_docs_from_collection(rc.client, "presentations"), key=_id_key)
        gtx["proprevs"] = sorted(all_docs_from_collection(rc.client, "proposalReviews"), key=_id_key)
        gtx["manrevs"] = sorted(all_docs_from_collection(rc.client, "refereeReports"), key=_id_key)
        gtx["recletts"] = sorted(all_docs_from_collection(rc.client, "recletts"), key=_id_key)
        gtx["patents"] = sorted(all_docs_from_collection(rc.client, "patents"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def latex(self):
        """Render latex template"""
        rc = self.rc
        group = fuzzy_retrieval(self.gtx["groups"], ["_id", "aka", "name"], rc.groupname)
        if not rc.people:
            raise RuntimeError("ERROR: please rerun specifying --people name")
        if not rc.from_date:
            raise RuntimeError("ERROR: please rerun specifying --from")
        build_target = get_id_from_name(all_docs_from_collection(rc.client, "people"), rc.people[0])
        begin_year = int(rc.from_date.split("-")[0])
        begin_period = date_parser.parse(rc.from_date).date()
        pre_begin_period = begin_period - relativedelta(years=1)
        if rc.to_date:
            to_date = date_parser.parse(rc.to_date).date()
            end_period = to_date
            post_end_period = to_date + relativedelta(years=1)
        else:
            end_period = begin_period + relativedelta(years=1) - relativedelta(days=1)
            post_end_period = begin_period + relativedelta(years=2) - relativedelta(days=1)

        me = [p for p in self.gtx["people"] if p["_id"] == build_target][0]
        me["begin_period"] = dt.date.strftime(begin_period, "%m/%d/%Y")
        me["begin_period"] = dt.date.strftime(begin_period, "%m/%d/%Y")
        me["pre_begin_period"] = dt.date.strftime(pre_begin_period, "%m/%d/%Y")
        me["end_period"] = dt.date.strftime(end_period, "%m/%d/%Y")
        me["post_end_period"] = dt.date.strftime(post_end_period, "%m/%d/%Y")
        projs = filter_projects(self.gtx["projects"], set([build_target]), group=group["_id"])
        ########
        # Recommendation Letters count
        ########
        recletts = self.gtx["recletts"]
        num_recletts = len(
            [reclett["_id"] for reclett in recletts if get_dates(reclett).get("end_date") >= begin_period]
        )
        ########
        # Proposal review count
        ########
        proprevs = self.gtx["proprevs"]
        num_proprevs = len(
            [
                proprev["_id"]
                for proprev in proprevs
                if get_dates(proprev).get("end_date") >= begin_period and proprev.get("status") == "submitted"
            ]
        )
        ########
        # Manuscript review count
        ########
        manrevs = self.gtx["manrevs"]
        num_manrevs = len(
            [
                manrev["_id"]
                for manrev in manrevs
                if manrev.get("status") == "submitted"
                and get_dates(manrev, date_field_prefix="submitted").get("submitted_date", dt.date(1971, 1, 1))
                is not None
                and get_dates(manrev, date_field_prefix="submitted").get("submitted_date", dt.date(1971, 1, 1))
                >= begin_period
            ]
        )

        #########
        # current and pending
        #########
        pi = fuzzy_retrieval(self.gtx["people"], ["_id", "aka", "name"], build_target)
        pinames = pi["name"].split()
        piinitialslist = [i[0] for i in pinames]
        pi["initials"] = "".join(piinitialslist).upper()

        grants = merge_collections_all(self.gtx["proposals"], self.gtx["grants"], "proposal_id")
        for g in grants:
            g["end_date"] = get_dates(g).get("end_date")
            g["begin_date"] = get_dates(g).get("begin_date", dt.date(1900, 1, 2))
            g["award_start_date"] = "{}/{}/{}".format(
                g.get("begin_date").month,
                g.get("begin_date").day,
                g.get("begin_date").year,
            )
            g["award_end_date"] = "{}/{}/{}".format(
                g.get("end_date").month, g.get("end_date").day, g.get("end_date").year
            )

            for person in g.get("team", []):
                rperson = fuzzy_retrieval(self.gtx["people"], ["aka", "name"], person["name"])
                if rperson:
                    person["name"] = rperson["name"]
            if g.get("budget"):
                amounts = [i.get("amount") for i in g.get("budget")]
                g["subaward_amount"] = sum(amounts)

        current_grants = [dict(g) for g in grants if is_current(g)]
        current_grants, _, _ = filter_grants(current_grants, {pi["name"]}, pi=False, multi_pi=True)
        current_grants = [g for g in current_grants if g.get("status") != "declined"]
        for g in current_grants:
            if g.get("budget"):
                amounts = [i.get("amount") for i in g.get("budget")]
                g["subaward_amount"] = sum(amounts)

        pending_grants = [g for g in self.gtx["proposals"] if is_pending(g["status"])]
        for g in pending_grants:
            for person in g["team"]:
                rperson = fuzzy_retrieval(self.gtx["people"], ["aka", "name"], person["name"])
                if rperson:
                    person["name"] = rperson["name"]
        pending_grants, _, _ = filter_grants(pending_grants, {pi["name"]}, pi=False, multi_pi=True)
        badids = [i["_id"] for i in current_grants if not i.get("cpp_info").get("cppflag", "")]

        declined_proposals = [g for g in self.gtx["proposals"] if is_declined(g["status"])]
        for g in declined_proposals:
            for person in g["team"]:
                rperson = fuzzy_retrieval(self.gtx["people"], ["aka", "name"], person["name"])
                if rperson:
                    person["name"] = rperson["name"]
        declined_proposals, _, _ = filter_grants(declined_proposals, {pi["name"]}, pi=False, multi_pi=True)
        declined_proposals = [
            proposal
            for proposal in declined_proposals
            if get_dates(proposal).get("begin_date") >= begin_period
            and get_dates(proposal, date_field_prefix="submitted").get("submitted_date", end_period) <= end_period
        ]

        iter = copy(current_grants)
        for grant in iter:
            if grant["_id"] in badids:
                current_grants.remove(grant)
        #########
        # end current and pending
        #########

        #########
        # highlights
        #########
        ossoftware = False
        for proj in projs:
            if proj.get("highlights"):
                proj["current_highlights"] = False
                for highlight in proj.get("highlights"):
                    highlight_date = get_dates(highlight)
                    if highlight_date.get("end_date") >= begin_period:
                        highlight["is_current"] = True
                        proj["current_highlights"] = True
                        if proj.get("type") == "ossoftware":
                            ossoftware = True

        #########
        # advising
        #########
        undergrads = filter_employment_for_advisees(self.gtx["people"], begin_period, "undergrad", rc.people[0])
        masters = filter_employment_for_advisees(self.gtx["people"], begin_period, "ms", rc.people[0])
        currents = filter_employment_for_advisees(self.gtx["people"], begin_period, "phd", rc.people[0])
        graduateds = filter_employment_for_advisees(
            self.gtx["people"], begin_period.replace(year=begin_year - 5), "phd", rc.people[0]
        )
        postdocs = filter_employment_for_advisees(self.gtx["people"], begin_period, "postdoc", rc.people[0])
        visitors = filter_employment_for_advisees(
            self.gtx["people"], begin_period, "visitor-unsupported", rc.people[0]
        )
        iter = deepcopy(graduateds)
        for g in iter:
            if g.get("active"):
                graduateds.remove(g)
        iter = deepcopy(currents)
        for g in iter:
            if not g.get("active"):
                currents.remove(g)

        ######################
        # service
        #####################
        mego = deepcopy(me)
        dept_service = filter_service(mego, begin_period, "department")
        mego = deepcopy(me)
        school_service = filter_service(mego, begin_period, "school")
        mego = deepcopy(me)
        uni_service = filter_service(mego, begin_period, "university")
        uni_service.extend(school_service)
        if num_recletts > 0:
            uni_service.append({"name": f"Wrote recommendation letters for {num_recletts} " f"people this period"})
        mego = deepcopy(me)
        prof_service = filter_service(mego, begin_period, "profession")
        if num_proprevs > 0:
            prof_service.append(
                {"name": f"Reviewed {num_proprevs} funding proposals for " f"national agencies this period"}
            )
        if num_manrevs > 0:
            prof_service.append(
                {"name": f"Reviewed {num_manrevs} manuscripts for " f"peer reviewed journals this period"}
            )
        mego = deepcopy(me)
        phd_defenses = filter_committees(mego, begin_period, "phddefense")
        phd_proposals = filter_committees(mego, begin_period, "phdproposal")
        phd_orals = filter_committees(mego, begin_period, "phdoral")
        mego = deepcopy(me)
        outreach = filter_service(mego, begin_period, "outreach")
        mego = deepcopy(me)
        lab = filter_facilities([mego], begin_period, "research")
        mego = deepcopy(me)
        shared = filter_facilities([mego], begin_period, "shared")
        mego = deepcopy(me)
        fac_other = filter_facilities([mego], begin_period, "other")
        mego = deepcopy(me)
        fac_teaching = filter_facilities([mego], begin_period, "teaching")
        mego = deepcopy(me)
        fac_wishlist = filter_facilities([mego], begin_period, "research_wish", verbose=False)
        mego = deepcopy(me)
        tch_wishlist = filter_facilities([mego], begin_period, "teaching_wish")
        mego = deepcopy(me)
        curric_dev = filter_activities([mego], begin_period, "teaching")
        mego = deepcopy(me)
        other_activities = filter_activities([mego], begin_period, "other")

        ##########################
        # Presentation list
        ##########################
        keypres = filter_presentations(
            self.gtx["people"],
            self.gtx["presentations"],
            self.gtx["institutions"],
            build_target,
            types=["award", "plenary", "keynote"],
            since=begin_period,
            before=end_period,
            statuses=["accepted"],
        )
        invpres = filter_presentations(
            self.gtx["people"],
            self.gtx["presentations"],
            self.gtx["institutions"],
            build_target,
            types=["invited"],
            since=begin_period,
            before=end_period,
            statuses=["accepted"],
        )
        sempres = filter_presentations(
            self.gtx["people"],
            self.gtx["presentations"],
            self.gtx["institutions"],
            build_target,
            types=["colloquium", "seminar"],
            since=begin_period,
            before=end_period,
            statuses=["accepted"],
        )
        declpres = filter_presentations(
            self.gtx["people"],
            self.gtx["presentations"],
            self.gtx["institutions"],
            build_target,
            types=["all"],
            since=begin_period,
            before=end_period,
            statuses=["declined"],
        )

        #########################
        # Awards
        #########################
        ahs = awards(me, since=begin_period)

        ########################
        # Publications
        ########################
        names = frozenset(me.get("aka", []) + [me["name"]])
        pubs = filter_publications(
            all_docs_from_collection(rc.client, "citations"), names, reverse=True, bold=False, since=begin_period
        )
        # remove unpublished papers
        # unpubs = [pub for pub in pubs if len(pub.get("doi") == 0)]
        pubed = [pub for pub in pubs if len(pub.get("doi", "")) > 0]
        non_arts = [pub for pub in pubs if pub.get("entrytype") != "article"]
        pubs = pubed + non_arts
        bibfile = make_bibtex_file(pubs, pid=me["_id"], person_dir=self.bldir)
        articles = [prc for prc in pubs if prc.get("entrytype") == "article" and not prc.get("peer_rev_conf")]
        NONARTICLETYPES = [
            "book",
            "inbook",
            "proceedings",
            "inproceedings",
            "incollection",
            "unpublished",
            "phdthesis",
            "misc",
        ]
        nonarticles = [prc for prc in pubs if prc.get("entrytype") in NONARTICLETYPES]
        peer_rev_conf_pubs = [prc for prc in pubs if prc.get("peer_rev_conf")]

        ##############
        # TODO: add Current Projects to Research summary section
        ##############

        #############
        # IP
        #############
        patents = filter_patents(self.gtx["patents"], self.gtx["people"], build_target, since=begin_period)
        licenses = filter_licenses(self.gtx["patents"], self.gtx["people"], build_target, since=begin_period)
        #############
        # hindex
        #############
        if not me.get("miscellaneous"):
            me["miscellaneous"] = {"metrics_for_success": []}
        if me.get("hindex"):
            hindex = sorted(me["hindex"], key=doc_date_key).pop()
        #########################
        # render
        #########################
        self.render(
            "columbia_annual_report.tex",
            f"{pi['_id']}-ann-report.tex",
            pi=pi,
            p=me,
            projects=projs,
            pending=pending_grants,
            current=current_grants,
            declined=declined_proposals,
            undergrads=undergrads,
            masters=masters,
            currentphds=currents,
            graduatedphds=graduateds,
            postdocs=postdocs,
            visitors=visitors,
            ossoftware=ossoftware,
            phd_defenses=phd_defenses,
            phd_proposals=phd_proposals,
            phd_orals=phd_orals,
            dept_service=dept_service,
            uni_service=uni_service,
            prof_service=prof_service,
            outreach=outreach,
            lab=lab,
            shared=shared,
            facilities_other=fac_other,
            fac_teaching=fac_teaching,
            fac_wishlist=fac_wishlist,
            tch_wishlist=tch_wishlist,
            curric_dev=curric_dev,
            other_activities=other_activities,
            keypres=keypres,
            invpres=invpres,
            sempres=sempres,
            declpres=declpres,
            sentencecase=sentencecase,
            monthstyle=month_fullnames,
            ahs=ahs,
            pubs=articles,
            nonarticles=nonarticles,
            peer_rev_conf_pubs=peer_rev_conf_pubs,
            bibfile=bibfile,
            patents=patents,
            licenses=licenses,
            hindex=hindex,
        )
        self.pdf(f"{pi['_id']}-ann-report.tex")



"""Builder for manuscript reviews."""

import sys

import nameparser
from dateutil import parser as dateparser
from gooey import GooeyParser

from regolith.dates import month_to_str_int
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import all_docs_from_collection

ALLOWED_STATI = ["invited", "accepted", "declined", "downloaded", "inprogress", "submitted", "cancelled"]


def subparser(subpi):
    date_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"

    subpi.add_argument(
        "name",
        help="Full name, or last name, of the first author",
    )
    subpi.add_argument("due_date", help="Due date", default="", **date_kwargs)
    subpi.add_argument("journal", help="journal to be published in", default="")
    subpi.add_argument("title", help="the title of the Manuscript", default="")
    subpi.add_argument("-q", "--requester", help="name, or id in contacts, of the editor requesting the review")
    subpi.add_argument("-s", "--status", choices=ALLOWED_STATI, help="Manuscript status", default="accepted")
    subpi.add_argument("-d", "--submitted-date", help="Submitted date. Defaults " "to tbd", **date_kwargs)
    subpi.add_argument("-r", "--reviewer", help="name of the reviewer. Defaults to the one saved in user.json. ")
    subpi.add_argument(
        "-t",
        "--database",
        help="The database that will be updated. Defaults to " "first database in the regolithrc.json file.",
    )
    return subpi


class ManuRevAdderHelper(DbHelperBase):
    btype = "a_manurev"
    needed_colls = ["refereeReports"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        rc.coll = "refereeReports"
        gtx["refereeReports"] = sorted(all_docs_from_collection(rc.client, "refereeReports"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        name = nameparser.HumanName(rc.name)
        due_date = dateparser.parse(rc.due_date).date()
        if name.last == "":
            key = "{}{}_{}".format(
                str(due_date.year)[-2:], month_to_str_int(due_date.month), name.first.casefold().strip(".")
            )
        else:
            key = "{}{}_{}_{}".format(
                str(due_date.year)[-2:],
                month_to_str_int(due_date.month),
                name.last.casefold(),
                name.first.casefold().strip("."),
            )
        coll = self.gtx[rc.coll]
        pdocl = list(filter(lambda doc: doc["_id"] == key, coll))
        if len(pdocl) > 0:
            sys.exit("This entry appears to already exist in the collection")
        else:
            pdoc = {}
        pdoc.update(
            {
                "claimed_found_what": [],
                "claimed_why_important": [],
                "did_how": [],
                "did_what": [],
                "due_date": due_date,
                "editor_eyes_only": "",
                "final_assessment": [],
                "freewrite": "",
                "journal": rc.journal,
                "recommendation": "",
                "title": rc.title,
                "validity_assessment": [],
                "year": due_date.year,
            }
        )

        if rc.reviewer:
            pdoc.update({"reviewer": rc.reviewer})
        else:
            try:
                rc.reviewer = rc.default_user_id
                pdoc.update({"reviewer": rc.reviewer})
            except AttributeError:
                print(
                    "Please set default_user_id in '~/.config/regolith/user.json', "
                    "or you need to enter your group id in the command line"
                )
                return
        if rc.submitted_date:
            pdoc.update({"submitted_date": rc.submitted_date})
        else:
            pdoc.update({"submitted_date": "tbd"})
        if rc.name:
            if name.last == "":
                pdoc.update({"first_author_last_name": name.first})
            else:
                pdoc.update({"first_author_last_name": name.last})
        if rc.requester:
            pdoc.update({"requester": rc.requester})
        else:
            pdoc.update({"requester": ""})
        if rc.status:
            if rc.status not in ALLOWED_STATI:
                raise ValueError("status should be one of {}".format(ALLOWED_STATI))
            else:
                pdoc.update({"status": rc.status})
        else:
            pdoc.update({"status": "accepted"})

        pdoc.update({"_id": key})
        rc.client.insert_one(rc.database, rc.coll, pdoc)

        print("{} manuscript has been added/updated in manuscript reviews".format(rc.name))

        return



"""Helper for listing filtered data from collections in the database."""

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.tools import all_docs_from_collection, collection_str, get_pi_id, search_collection

HELPER_TARGET = "lister"


def subparser(subpi):
    subpi.add_argument("coll", help="Collection that the lister is going to list from.")
    subpi.add_argument(
        "-f",
        "--kv-filter",
        nargs="+",
        help="Search the given collection by key-value pairs. "
        "e.g. 'regolith helper lister -f name simon' will "
        "return the id of all the people who's name contains simon.",
    )
    subpi.add_argument(
        "-r",
        "--return-fields",
        nargs="+",
        help="Specify from which keys to print values. "
        "e.g. 'regolith helper lister people -r name status title' "
        "will return the name, status, and title of all the people "
        "in the collection. If no argument is given the default is just the id.",
    )
    subpi.add_argument("-k", "--keys", action="store_true", help="List of the available keys in the collection.")
    return subpi


class GeneralListerHelper(SoutHelperBase):
    """Helper for listing filtered data from collections in the database."""

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        needed_colls = [rc.coll]
        if "groups" in needed_colls:
            rc.pi_id = get_pi_id(rc)
        colls = [sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in needed_colls]
        for db, coll in zip(needed_colls, colls):
            gtx[db] = coll
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc
        coll = self.gtx[rc.coll]
        keys = sorted(set([i for k in coll for i in k.keys()]))
        list_keys = []
        if rc.kv_filter:
            list_keys.extend([rc.kv_filter[n] for n in range(len(rc.kv_filter)) if n % 2 == 0])
        if rc.return_fields:
            list_keys.extend(rc.return_fields)
        for i in list_keys:
            if i not in keys:
                raise ValueError(f"{i} is not a valid key. Please choose a valid key from: {keys}")
        if len(coll) == 0:
            raise RuntimeError("This collection is empty or does not exist. Please inform a valid collection.")
        if rc.kv_filter:
            filter_result = search_collection(coll, rc.kv_filter).strip()
            if filter_result == "":
                print("There are no results that match your search.")
            else:
                if rc.return_fields:
                    print(
                        "Results of your search:\n"
                        f"{(search_collection(coll, rc.kv_filter, rc.return_fields).strip())}"
                    )
                else:
                    print(f"Results of your search:\n{filter_result}")
        if rc.return_fields and not rc.kv_filter:
            print(f"Results of your search:\n{collection_str(coll, rc.return_fields).strip()}")
        if rc.keys:
            print(f"Available keys:\n{keys}")
            return
        if not rc.kv_filter and not rc.return_fields:
            print(f"Results of your search:\n{collection_str(coll).strip()}")
        return



"""Generic builder."""

from copy import copy

from regolith.helpers import a_expensehelper as a_expense
from regolith.helpers import a_grppub_readlisthelper as a_gprl
from regolith.helpers import a_manurevhelper as a_manurev
from regolith.helpers import a_presentationhelper as a_presentation
from regolith.helpers import a_projectumhelper as a_projectum
from regolith.helpers import a_proposalhelper as a_proposal
from regolith.helpers import a_proprevhelper as a_proprev
from regolith.helpers import a_todohelper as a_todo
from regolith.helpers import attestationshelper as attestations
from regolith.helpers import f_todohelper as f_todo
from regolith.helpers import l_abstracthelper as l_abstract
from regolith.helpers import l_contactshelper as l_contacts
from regolith.helpers import l_currentappointmentshelper as l_currentappointments
from regolith.helpers import l_generalhelper as l_general
from regolith.helpers import l_grantshelper as l_grants
from regolith.helpers import l_membershelper as l_members
from regolith.helpers import l_milestoneshelper as l_milestone
from regolith.helpers import l_progressreporthelper as l_progress
from regolith.helpers import l_projectahelper as l_projecta
from regolith.helpers import l_todohelper as l_todo
from regolith.helpers import makeappointmentshelper as makeappointments
from regolith.helpers import reimbstatushelper as reimbstatus
from regolith.helpers import u_contacthelper as u_contact
from regolith.helpers import u_finishprumhelper as u_finishprum
from regolith.helpers import u_institutionshelper as u_institutions
from regolith.helpers import u_logurlhelper as u_logurl
from regolith.helpers import u_milestonehelper as u_milestone
from regolith.helpers import u_todohelper as u_todo
from regolith.helpers import v_meetingshelper as v_meetings

# Updtaer helpers will update the db and should not load all databases but only
# the one specified in rc.database for updating.
UPDATER_HELPERS = {
    "a_expense": (a_expense.ExpenseAdderHelper, a_expense.subparser),
    "a_grppub_readlist": (a_gprl.GrpPubReadListAdderHelper, a_gprl.subparser),
    "a_manurev": (a_manurev.ManuRevAdderHelper, a_manurev.subparser),
    "a_presentation": (a_presentation.PresentationAdderHelper, a_presentation.subparser),
    "a_projectum": (a_projectum.ProjectumAdderHelper, a_projectum.subparser),
    "a_proposal": (a_proposal.ProposalAdderHelper, a_proposal.subparser),
    "a_proprev": (a_proprev.PropRevAdderHelper, a_proprev.subparser),
    "a_todo": (a_todo.TodoAdderHelper, a_todo.subparser),
    "f_prum": (u_finishprum.FinishprumUpdaterHelper, u_finishprum.subparser),
    "f_todo": (f_todo.TodoFinisherHelper, f_todo.subparser),
    "u_contact": (u_contact.ContactUpdaterHelper, u_contact.subparser),
    "u_institution": (u_institutions.InstitutionsUpdaterHelper, u_institutions.subparser),
    "u_logurl": (u_logurl.LogUrlUpdaterHelper, u_logurl.subparser),
    "u_milestone": (u_milestone.MilestoneUpdaterHelper, u_milestone.subparser),
    "u_todo": (u_todo.TodoUpdaterHelper, u_todo.subparser),
}

# Lister helpers need to load collections across all the databses to show everything
LISTER_HELPERS = {
    "l_abstract": (l_abstract.AbstractListerHelper, l_abstract.subparser),
    "l_contacts": (l_contacts.ContactsListerHelper, l_contacts.subparser),
    "l_currentappointments": (
        l_currentappointments.CurrentAppointmentsListerHelper,
        l_currentappointments.subparser,
    ),
    "l_grants": (l_grants.GrantsListerHelper, l_grants.subparser),
    "l_members": (l_members.MembersListerHelper, l_members.subparser),
    "l_milestones": (l_milestone.MilestonesListerHelper, l_milestone.subparser),
    "l_progress": (l_progress.ProgressReportHelper, l_progress.subparser),
    "l_projecta": (l_projecta.ProjectaListerHelper, l_projecta.subparser),
    "l_reimbstatus": (reimbstatus.ReimbstatusHelper, reimbstatus.subparser),
    "l_todo": (l_todo.TodoListerHelper, l_todo.subparser),
    "v_meetings": (v_meetings.MeetingsValidatorHelper, v_meetings.subparser),
    "attestations": (attestations.AttestationsHelper, attestations.subparser),
    "lister": (l_general.GeneralListerHelper, l_general.subparser),
    "makeappointments": (makeappointments.MakeAppointmentsHelper, makeappointments.subparser),
}

HELPERS = copy(LISTER_HELPERS)
HELPERS.update(UPDATER_HELPERS)
# fast_updater updaters only connects to the one requested db, not to all dbs
# in rc.databases which is the default behavior
FAST_UPDATER_WHITELIST = ["u_milestone", "f_prum"]


def helpr(btype, rc):
    """Returns helper of the appropriate type."""
    return HELPERS[btype][0](rc)



"""
Helper for updating/adding  to the projecta collection.
"""

import datetime as dt
import uuid

from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import all_docs_from_collection, fragment_retrieval

TARGET_COLL = "institutions"


def subparser(subpi):
    date_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"

    subpi.add_argument(
        "institution_id",
        help="id of the institution, e.g., columbiau. "
        "If an index is not specified, this will return "
        "a numbered list of all institutions that contain this id "
        "fragment.  Specify -i # to update institution number # "
        "in the list. #1 is always used for a new institution.",
    )
    subpi.add_argument("-i", "--index", help="Index of the item in the enumerated list to update.", type=int)
    subpi.add_argument("-n", "--name", help="Full name of the institution")
    subpi.add_argument("--city", help="The city where the institution is. " "Required for a new institution.")
    subpi.add_argument(
        "--state",
        help="The state where the institution is. "
        "Required for a new institution if institution's country is US.",
    )
    subpi.add_argument(
        "--zip",
        help="zipcode of the institution. " "Required for a new institution if institution's country is US.",
    )
    subpi.add_argument("--country", help="The country where the institution is. " "Required for a new institution")
    subpi.add_argument(
        "-a", "--aka", nargs="+", help="List of all the different names this " "institution is known by."
    )
    subpi.add_argument("--dept-id", help="dept_id, e.g. physics.")
    subpi.add_argument(
        "--dept-name",
        help="Department canonical name, e.g., Department of Physics. "
        "Required if --dept-id supplied and it is a new department",
    )
    subpi.add_argument("--dept-aka", nargs="+", help="Department aliases, e.g., Physics Dept.")
    subpi.add_argument("--school-id", help="id for the school, e.g., SEAS.")
    subpi.add_argument(
        "--school-name",
        help="Full canonical name, e.g., School of Engineering and Applied Science. "
        "Required if --school-id supplied and it is a new school",
    )
    subpi.add_argument("--school-aka", nargs="+", help="School aliases.")
    # Do not delete --database arg
    subpi.add_argument(
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    # Do not delete --date arg
    subpi.add_argument(
        "--date", help="The date when the institution was created. " "Defaults to today's date.", **date_kwargs
    )
    return subpi


class InstitutionsUpdaterHelper(DbHelperBase):
    """
    Helper for updating/adding  to the projecta collection.
    """

    # btype must be the same as helper target in helper.py
    btype = "u_institution"
    needed_colls = [f"{TARGET_COLL}"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        rc.coll = f"{TARGET_COLL}"
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        key = rc.institution_id
        filterid = {"_id": key}
        target_inst = rc.client.find_one(rc.database, rc.coll, filterid)
        now = dt.datetime.today()
        pdoc = {}
        departments = {}
        schools = {}
        if target_inst:
            if rc.aka:
                current_aka = target_inst.get("aka")
                current_aka.extend(rc.aka)
                pdoc.update({"aka": current_aka})
            departments = target_inst.get("departments", {})
            schools = target_inst.get("schools", {})
        else:
            inst = fragment_retrieval(self.gtx["institutions"], ["_id", "name", "aka"], rc.institution_id)
            inst.sort(key=lambda x: x["_id"], reverse=False)
            if not rc.index:
                print("Please rerun the helper specifying '-n list-index' to update item number 'list-index':")
                print(f"1. {key} as a new institution.")
                for i in range(len(inst)):
                    print(f"{i+2}. {inst[i].get('_id')}      {inst[i].get('name')}.")
                return
            if rc.index < 1 or rc.index > len(inst) + 1:
                raise RuntimeError("Sorry, you picked an invalid number.")
            if rc.index == 1:
                if not rc.name or not rc.city or not rc.country:
                    raise RuntimeError("Name, city, and country are required for a new institution.")
                if rc.country == "US":
                    if not rc.zip or not rc.state:
                        raise RuntimeError(
                            "Zip and state are required for a new institution " "if institutions is in the US."
                        )
                pdoc.update({"name": rc.name, "uuid": str(uuid.uuid4())})
                if rc.aka:
                    pdoc.update({"aka": rc.aka})
                if rc.date:
                    pdoc.update({"date": rc.date})
                else:
                    pdoc.update({"date": now.date()})
            else:
                chosen_inst = inst[rc.index - 2]
                key = chosen_inst.get("_id")
                if rc.aka:
                    current_aka = chosen_inst.get("aka")
                    current_aka.extend(rc.aka)
                    pdoc.update({"aka": current_aka})
                current_departments = chosen_inst.get("departments")
                for k, v in current_departments.items():
                    info_department = {"name": v.get("name"), "aka": v.get("aka")}
                    departments.update({k: info_department})
                current_schools = chosen_inst.get("schools")
                for k, v in current_schools.items():
                    info_school = {"name": v.get("name"), "aka": v.get("aka")}
                    schools.update({k: info_school})
        if rc.city:
            pdoc.update({"city": rc.city})
        if rc.state:
            pdoc.update({"state": rc.state})
        if rc.country:
            pdoc.update({"country": rc.country})
        if rc.zip:
            pdoc.update({"zip": rc.zip})
        if rc.date:
            pdoc.update({"updated": rc.date})
        else:
            pdoc.update({"updated": now})
        # departments:
        if rc.dept_id:
            dep = {}
            if rc.dept_id in departments:
                doc = departments.get(rc.dept_id)
                current_dept_aka = doc.get("aka")
                if rc.dept_name:
                    dep.update({"name": rc.dept_name})
                if rc.dept_aka:
                    current_dept_aka.extend(rc.dept_aka)
                    doc.update({"aka": current_dept_aka})
                departments.update({rc.dept_id: doc})
            else:
                if not rc.dept_name:
                    raise RuntimeError("Name is required for a new department.")
                dep.update({"name": rc.dept_name})
                if rc.dept_aka:
                    dep.update({"aka": rc.dept_aka})
                departments[rc.dept_id] = dep
            pdoc.update({"departments": departments})
        # schools
        if rc.school_id:
            school = {}
            if rc.school_id in schools:
                doc = schools.get(rc.school_id)
                if rc.school_name:
                    doc.update({"name": rc.school_name})
                current_sc_aka = doc.get("aka")
                if rc.dept_aka:
                    current_sc_aka.extend(rc.school_aka)
                    doc.update({"aka": current_sc_aka})
                schools.update({rc.school_id: doc})
            else:
                if not rc.school_name:
                    raise RuntimeError("Name is required for a new school.")
                school.update({"name": rc.school_name})
                if rc.school_aka:
                    school.update({"aka": rc.school_aka})
                schools[rc.school_id] = school
                pdoc.update({"schools": schools})
        rc.client.update_one(rc.database, rc.coll, {"_id": key}, pdoc)
        print(f"{key} has been updated/added in institutions")
        return



import copy
import os
import subprocess
import sys

import pytest

from regolith.runcontrol import DEFAULT_RC, load_rcfile
from regolith.storage import find_store, storage_path


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
def test_cmd(make_db, tmpdir):
    repo = make_db
    os.chdir(repo)
    subprocess.check_call(["touch", "myfile2.tex"], cwd=tmpdir)
    subprocess.check_call(["regolith", "store", "store", os.path.join(tmpdir, "myfile2.tex")])
    rc = copy.copy(DEFAULT_RC)
    rc._update(load_rcfile("regolithrc.json"))
    assert "myfile.tex" in os.listdir(storage_path(find_store(rc), rc))



"""
Loads the dbs for interactive sessions
"""

from regolith.runcontrol import DEFAULT_RC, connect_db, filter_databases, load_rcfile

rc = DEFAULT_RC
rc._update(load_rcfile("regolithrc.json"))
filter_databases(rc)

chained_db, dbs = connect_db(rc)



"""Helper for finishing prum in the projecta collection"""

import datetime as dt

from dateutil import parser as date_parser
from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import all_docs_from_collection, fragment_retrieval

TARGET_COLL = "projecta"


def subparser(subpi):
    date_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"

    subpi.add_argument(
        "projectum_id", help="the ID or fragment of the ID of the projectum to be updated, e.g., 20sb"
    )
    subpi.add_argument("--end-date", help="End date of the projectum. " "Defaults to today.", **date_kwargs)
    # Do not delete --database arg
    subpi.add_argument(
        "-d",
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    return subpi


class FinishprumUpdaterHelper(DbHelperBase):
    """
    Helper for finishing prum in the projecta collection
    """

    # btype must be the same as helper target in helper.py
    btype = "f_prum"
    needed_colls = [f"{TARGET_COLL}"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        rc.coll = f"{TARGET_COLL}"
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        key = rc.projectum_id
        filterid = {"_id": key}
        found_projectum = rc.client.find_one(rc.database, rc.coll, filterid)
        if not found_projectum:
            pra = fragment_retrieval(self.gtx["projecta"], ["_id"], key)
            if len(pra) == 0:
                raise RuntimeError("Please input a valid projectum id or a valid fragment of a projectum id")
            print("Projectum not found. Projecta with similar names: ")
            for i in range(len(pra)):
                print(f"{pra[i].get('_id')}     status:{pra[i].get('status')}")
            print("Please rerun the helper specifying the complete ID.")
            return
        found_projectum.update({"status": "finished"})
        if rc.end_date:
            end_date = date_parser.parse(rc.end_date).date()
        else:
            end_date = dt.date.today()
        found_projectum.update({"end_date": end_date})
        if found_projectum.get("kickoff"):
            if not found_projectum.get("kickoff", {}).get("end_date"):
                found_projectum["kickoff"].update({"status": "finished", "end_date": end_date})
            else:
                found_projectum["kickoff"].update({"status": "finished"})
        found_projectum["deliverable"].update({"status": "finished"})
        for milestone in found_projectum.get("milestones"):
            if not milestone.get("end_date"):
                milestone.update({"status": "finished", "end_date": end_date})
            else:
                milestone.update({"status": "finished"})
        rc.client.update_one(rc.database, rc.coll, filterid, found_projectum)
        print(f"{rc.projectum_id} status has been updated to finished")
        return



"""Classlist implementation"""

import csv
import json
import os
import re
import sys
from html.parser import HTMLParser
from pprint import pformat, pprint

from nameparser import HumanName


def load_json(filename):
    """Returns students as a list of dicts from JSON file."""
    with open(filename, encoding="utf-8") as f:
        students = json.load(f)
    return students


def load_csv(filename, format="columbia"):
    """Returns students as a list of dicts from a csv from Columbia Courseworks"""
    with open(filename, encoding="utf-8") as f:
        reader = csv.DictReader(f)
        students = []
        for row in reader:
            students.append(row)

    if format == "columbia":
        email_suffix = "@columbia.edu"
        [
            stud.update(
                {
                    "first_name": HumanName(stud["Student"]).first,
                    "last_name": HumanName(stud["Student"]).last,
                    "email": f"{stud.get('SIS User ID').strip()}{email_suffix}",
                    "university_id": stud.get("SIS User ID"),
                }
            )
            for stud in students
        ]

    for student in students:
        student["first_name"] = student.get("first_name").strip()
        student["last_name"] = student.get("last_name").strip()
        student["_id"] = "{} {}".format(student.get("first_name"), student.get("last_name")).strip()
        student["email"] = student.get("email").strip()
        student["university_id"] = student["university_id"].strip()

    return students


RE_ID = re.compile(r"^[A-Z]\d+$")
RE_NAME = re.compile("^[A-Za-z-]+$")


def _check_name(name, label, full):
    parts = (name[:-1] if name.endswith(".") else name).split()
    for part in parts:
        if RE_NAME.match(part) is None:
            print("skipping because of {} name: {}".format(label, full))
            return False
    return True


class UscHtmlParser(HTMLParser):
    """Class for parsing data from USC-formatted HTML."""

    def __init__(self):
        super().__init__()
        self.student = None
        self.students = []
        self.intable = self.inrow = self.incol = False

    def should_handle(self):
        return self.intable or self.inrow or self.incol

    def handle_starttag(self, tag, attrs):
        if tag == "table":  # we don't support nesting tables
            self.intable = True
        if not self.should_handle():
            return
        elif tag == "tr":
            self.inrow = True
            self.student = {}
        elif tag == "td":
            self.incol = True
        elif tag == "a" and self.incol:
            for attr, val in attrs:
                if attr == "href" and val.startswith("mailto:"):
                    _, _, email = val.partition(":")
                    self.student["email"] = email
                    break

    def handle_endtag(self, tag):
        if not self.should_handle():
            return
        if tag == "table":
            self.intable = False
        elif tag == "tr":
            if len(self.student) > 0 and "_id" in self.student:
                self.students.append(self.student)
            self.student = None
            self.inrow = False
        elif tag == "td":
            self.incol = False

    def handle_data(self, data):
        if not self.incol:
            return
        if "," in data:
            # found name
            last, _, first = data.partition(",")
            first = first.strip()
            last = last.strip()
            if not _check_name(first, "first", data):
                return
            if not _check_name(last, "last", data):
                return
            self.student["_id"] = first + " " + last
        elif RE_ID.match(data) is not None:
            self.student["university_id"] = data


def load_usc(filename):
    """Returns students as a list of dicts from an HTML file obtainted from the University of
    South Carolina.
    """
    with open(filename, encoding="utf-8") as f:
        html = f.read()
    parser = UscHtmlParser()
    parser.feed(html)
    return parser.students


LOADERS = {"usc": load_usc, "json": load_json, "csv": load_csv}


def add_students_to_db(students, rc):
    """Add new students to the student directory."""
    for student in students:
        rc.client.update_one(rc.db, "students", {"_id": student["_id"]}, student, upsert=True)


def add_students_to_course(students, rc):
    """Add students to the course listed"""
    course = rc.client.find_one(rc.db, "courses", {"_id": rc.course_id})
    if not course:
        raise ValueError(f"no course {rc.course_id} found in database")
    registry = {s["_id"] for s in students}
    if rc.op == "add":
        registry |= set(course["students"])
    elif rc.op == "replace":
        pass
    else:
        raise ValueError("operation {0!r} nor recognized".format(rc.op))
    course["students"] = sorted(registry)
    rc.client.update_one(rc.db, "courses", {"_id": rc.course_id}, course, upsert=True)


def register(rc):
    """Entry point for registering classes."""
    if not os.path.exists(rc.filename):
        sys.exit(
            "classlist file {} can't be found\nPlease check the filename " "and try again".format(rc.filename)
        )
    if rc.format is None:
        rc.format = os.path.splitext(rc.filename)[1][1:]
    loader = LOADERS[rc.format]
    students = loader(rc.filename)
    if rc.dry_run:
        pprint(students)
        return
    if rc.db is None:
        dbs = rc.client.keys()
        if len(dbs) == 1:
            rc.db = list(dbs)[0]
        else:
            raise RuntimeError(
                "More than one database present in run control, "
                'please select one with the "--db" option. '
                "Available dbs are: " + pformat(dbs)
            )
    add_students_to_db(students, rc)
    add_students_to_course(students, rc)



"""Helper for listing the to-do tasks.

Tasks are gathered from people.yml, milestones, and group meeting actions."""

import datetime as dt
import math

import dateutil.parser as date_parser
from gooey import GooeyParser
from nameparser import HumanName

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.schemas import PROJECTUM_ACTIVE_STATI, alloweds
from regolith.tools import (
    all_docs_from_collection,
    document_by_value,
    get_pi_id,
    key_value_pair_filter,
    print_task,
)

TARGET_COLL = "todos"
HELPER_TARGET = "l_todo"
Importance = [3, 2, 1, 0, -1, -2]  # eisenhower matrix (important|urgent) tt=3, tf=2, ft=1, ff=0
STATI = ["accepted", "downloaded", "inprep"]

TODO_STATI = alloweds.get("TODO_STATI")


def subparser(subpi):
    listbox_kwargs = {}
    date_kwargs = {}
    int_kwargs = {}
    if isinstance(subpi, GooeyParser):
        listbox_kwargs["widget"] = "Listbox"
        date_kwargs["widget"] = "DateChooser"
        int_kwargs["widget"] = "IntegerField"
        int_kwargs["gooey_options"] = {"min": 0, "max": 1000}

    subpi.add_argument(
        "--short",
        nargs="?",
        const=30,
        help="Filter for tasks of short duration. "
        "All items with a duration <= # mins, will be returned "
        "if the number # is specified.",
    )
    subpi.add_argument(
        "-t",
        "--tags",
        nargs="+",
        help="Filter tasks by tags. Items are returned if they contain any of the tags listed",
    )
    subpi.add_argument(
        "-s",
        "--stati",
        nargs="+",
        choices=TODO_STATI,
        help="Filter tasks with specific stati",
        default=["started"],
        **listbox_kwargs,
    )
    subpi.add_argument("-o", "--outstandingreview", help="List outstanding reviews", action="store_true")
    subpi.add_argument(
        "-a",
        "--assigned-to",
        help="Filter tasks that are assigned to this user id. Default id is saved in user.json. ",
    )
    subpi.add_argument(
        "-b",
        "--assigned-by",
        nargs="?",
        const="default_id",
        help="Filter tasks that are assigned to other members by this user id. Default id is saved in user.json. ",
    )
    subpi.add_argument(
        "--date",
        help="Enter a date such that the helper can calculate how many days are left "
        "from that date to the due-date. Default is today.",
        **date_kwargs,
    )
    subpi.add_argument(
        "-f",
        "--filter",
        nargs="+",
        help="Search this collection by giving key element pairs. '-f description paper' will return tasks "
        "with description containing 'paper' ",
    )

    return subpi


class TodoListerHelper(SoutHelperBase):
    """Helper for listing the to-do tasks.

    Tasks are gathered from people.yml, milestones, and group meeting actions."""

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET
    needed_colls = [f"{TARGET_COLL}", "refereeReports", "proposalReviews"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        colls = [
            sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in self.needed_colls
        ]
        for db, coll in zip(self.needed_colls, colls):
            gtx[db] = coll
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc
        if not rc.assigned_to:
            try:
                rc.assigned_to = rc.default_user_id
            except AttributeError:
                print(
                    "Please set default_user_id in '~/.config/regolith/user.json', "
                    "or you need to enter your group id in the command line"
                )
                return
        try:
            person = document_by_value(all_docs_from_collection(rc.client, "todos"), "_id", rc.assigned_to)
            gather_todos = person.get("todos", [])
        except Exception:
            print("The id you entered can't be found in todos.yml.")
            return
        if not rc.date:
            today = dt.date.today()
        else:
            today = date_parser.parse(rc.date).date()
        if rc.stati == ["started"]:
            rc.stati = PROJECTUM_ACTIVE_STATI
        if rc.filter:
            gather_todos = key_value_pair_filter(gather_todos, rc.filter)
        if rc.short:
            for todo in gather_todos[::-1]:
                if todo.get("duration") is None or float(todo.get("duration")) > float(rc.short):
                    gather_todos.remove(todo)
        if rc.tags:
            for todo in gather_todos[::-1]:
                takeme = False
                for tag in rc.tags:
                    if tag in todo.get("tags", []):
                        takeme = True
                if not takeme:
                    gather_todos.remove(todo)
        if rc.assigned_by:
            if rc.assigned_by == "default_id":
                rc.assigned_by = rc.default_user_id
            for todo in gather_todos[::-1]:
                if todo.get("assigned_by") != rc.assigned_by:
                    gather_todos.remove(todo)
        len_of_started_tasks = 0
        milestones = 0
        for todo in gather_todos:
            if "milestone: " in todo["description"]:
                milestones += 1
            elif todo["status"] == "started":
                len_of_started_tasks += 1
        len_of_tasks = len(gather_todos)  # - milestones
        for todo in gather_todos:
            _format_todos(todo, today)
        gather_todos[:len_of_tasks] = sorted(
            gather_todos[:len_of_tasks],
            key=lambda k: (k["status"], k["importance"], k["order"], -k.get("duration", 10000)),
        )
        gather_todos[len_of_started_tasks:len_of_tasks] = sorted(
            gather_todos[len_of_started_tasks:len_of_tasks], key=lambda k: (-k["sort_finished"])
        )
        gather_todos[len_of_tasks:] = sorted(
            gather_todos[len_of_tasks:], key=lambda k: (k["status"], k["order"], -k.get("duration", 10000))
        )
        print(
            "If the indices are far from being in numerical order, please renumber them "
            "by running regolith helper u_todo -r"
        )
        print("(index) action (days to due date|importance|expected duration (mins)|tags|assigned by)")
        print("-" * 80)
        if len(gather_todos) != 0:
            print_task(gather_todos, stati=rc.stati)

        if rc.outstandingreview:
            prop = self.gtx["proposalReviews"]
            man = self.gtx["refereeReports"]
            outstanding_todo = []
            for manuscript in man:
                if manuscript.get("reviewer") != rc.assigned_to:
                    continue
                if manuscript.get("status") in STATI:
                    out = (
                        f"Manuscript by {manuscript.get('first_author_last_name')} in {manuscript.get('journal')} "
                        f"is due on {manuscript.get('due_date')}"
                    )
                    outstanding_todo.append((out, manuscript.get("due_date"), manuscript.get("status")))
            for proposal in prop:
                if proposal.get("reviewer") != rc.assigned_to:
                    continue
                if proposal.get("status") in STATI:
                    if isinstance(proposal.get("names"), str):
                        name = HumanName(proposal.get("names"))
                    else:
                        name = HumanName(proposal.get("names")[0])
                    out = (
                        f"Proposal by {name.last} for {proposal.get('agency')} ({proposal.get('requester')})"
                        f"is due on {proposal.get('due_date')}"
                    )
                    outstanding_todo.append((out, proposal.get("due_date"), proposal.get("status")))

            if len(outstanding_todo) != 0:
                print("-" * 30)
                print("Outstanding Reviews:")
                print("-" * 30)
            outstanding_todo = sorted(outstanding_todo, key=lambda k: str(k[1]))
            for stati in STATI:
                if stati in [output[2] for output in outstanding_todo]:
                    print(f"{stati}:")
                else:
                    continue
                for output in outstanding_todo:
                    if output[2] == stati:
                        print(output[0])

        return


def _format_todos(todo, today):
    """datify dates, set orders etc and update to-do items in place

    Parameters
    ----------
    todo: dict
      the to-do item to be munged

    Returns
    -------
    nothing

    """
    if isinstance(todo["due_date"], str):
        todo["due_date"] = date_parser.parse(todo["due_date"]).date()
    if isinstance(todo.get("end_date"), str):
        todo["end_date"] = date_parser.parse(todo["end_date"]).date()
    todo["days_to_due"] = (todo.get("due_date") - today).days
    todo["sort_finished"] = (todo.get("end_date", dt.date(1900, 1, 1)) - dt.date(1900, 1, 1)).days
    try:
        todo["order"] = 1 / (1 + math.exp(abs(todo["days_to_due"] - 0.5)))
    except OverflowError:
        todo["order"] = float("inf")
    return



import os
import subprocess
import tempfile

from regolith.broker import load_db

TMPDIR = tempfile.gettempdir()


def test_round_trip(make_db, tmpdir):
    repo = make_db
    os.chdir(repo)
    subprocess.check_call(["touch", "myfile.tex"], cwd=tmpdir)
    db = load_db()
    db.add_file(db["projects"]["Cyclus"], "myfile", os.path.join(tmpdir, "myfile.tex"))
    ret = db.get_file_path(db["projects"]["Cyclus"], "myfile")
    assert ret == os.path.join(repo, "myfile.tex")



"""Client interface for MongoDB.
Maintained such that only pymongo is necessary when using helper/builders, and additional command-line tools
are necessary to install for maintenance tasks, such as fs-to-mongo."""

import datetime
import itertools
import os
import shutil
import subprocess
import sys
import time
import urllib
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from tempfile import TemporaryDirectory

from ruamel.yaml import YAML

from regolith.tools import validate_doc

#
# setup mongo
#
try:
    import pymongo

    MONGO_AVAILABLE = True
except ImportError:
    print(
        "pymongo not found. Please install it following the instructions "
        "https://pymongo.readthedocs.io/en/stable/installation.html"
    )
    MONGO_AVAILABLE = False

from pymongo.collection import Collection

from regolith import fsclient
from regolith.tools import dbpathname, fallback

if not MONGO_AVAILABLE:
    ON_PYMONGO_V2 = ON_PYMONGO_V3 = False
elif pymongo.version.split(".")[0] == "2":
    ON_PYMONGO_V2 = True
    ON_PYMONGO_V3 = False
else:
    ON_PYMONGO_V2 = False
    ON_PYMONGO_V3 = True


def import_jsons(dbpath: str, dbname: str, host: str = None, uri: str = None) -> None:
    """Import the json files to mongo db.

    Each json file will be a collection in the database. The _id will be the same as it is in the json file.

    Parameters
    ----------
    dbpath : str
        The path to the db folder.

    dbname : str
        The name of the database in mongo.

    host : str
        The hostname or IP address or Unix domain socket path of a single mongod or mongos instance to connect
        to, or a mongodb URI, or a list of hostnames / mongodb URIs.

    uri : str
        Specify a resolvable URI connection string (enclose in quotes) to connect to the MongoDB deployment.
    """
    for json_path in Path(dbpath).glob("*.json"):
        cmd = ["mongoimport"]
        if host is not None:
            cmd += ["--host", host, "--db", dbname]
        if uri is not None:
            cmd += ["--uri", uri]
        cmd += ["--collection", json_path.stem, "--file", str(json_path)]
        try:
            subprocess.check_call(cmd, stderr=subprocess.STDOUT)
        except FileNotFoundError:
            print(
                "mongoimport command not found in environment path.\n\n"
                "If mongo server v4.4+ installed, download MongoDB Database Tools from:"
                " https://www.mongodb.com/try/download/database-tools\n"
                "and add C:\\Program Files\\MongoDB\\Tools\\<ToolsVersion>\\bin\\ to path.\n\n"
                "If mongo server <v4.4, ensure that C:\\Program Files\\MongoDB\\Server\\<ServerVersion>\\bin\\ \n"
                "has been added to the environment path.\n"
            )
            print("..................Upload failed..................")
        except subprocess.CalledProcessError as exc:
            print("Status : FAIL", exc.returncode, exc.output)
            raise exc
    return


def import_yamls(dbpath: str, dbname: str, host: str = None, uri: str = None) -> None:
    """Import the yaml files to mongo db.

    Each yaml file will be a collection in the database. The _id will be the id_key for each doc in the yaml file.

    Parameters
    ----------
    dbpath : str
        The path to the db folder.

    dbname : str
        The name of the database in mongo.

    host : str
        The hostname or IP address or Unix domain socket path of a single mongod or mongos instance to connect
        to, or a mongodb URI, or a list of hostnames / mongodb URIs.

    uri : str
        Specify a resolvable URI connection string (enclose in quotes) to connect to the MongoDB deployment.
    """
    yaml_files = itertools.chain(Path(dbpath).glob("*.yaml"), Path(dbpath).glob("*.yml"))
    with TemporaryDirectory() as tempd:
        for yaml_file in yaml_files:
            json_file = Path(tempd).joinpath(yaml_file.with_suffix(".json").name)
            loader = YAML(typ="safe")
            loader.constructor.yaml_constructors["tag:yaml.org,2002:timestamp"] = (
                loader.constructor.yaml_constructors["tag:yaml.org,2002:str"]
            )
            fsclient.yaml_to_json(str(yaml_file), str(json_file), loader=loader)
        import_jsons(tempd, dbname, host=host, uri=uri)
    return


def export_json(collection: str, dbpath: str, dbname: str, host: str = None, uri: str = None) -> None:
    cmd = ["mongoexport", "--collection", collection]
    if host is not None:
        cmd += ["--host", host, "--db", dbname]
    if uri is not None:
        cmd += ["--uri", uri]
    cmd += ["--out", str(os.path.join(dbpath, collection + ".json"))]
    try:
        subprocess.check_call(cmd, stderr=subprocess.STDOUT)
    except FileNotFoundError:
        print(
            "mongoexport command not found in environment path.\n\n"
            "If mongo server v4.4+ installed, download MongoDB Database Tools from:"
            " https://www.mongodb.com/try/download/database-tools\n"
            "and add C:\\Program Files\\MongoDB\\Tools\\<ToolsVersion>\\bin\\ to path.\n\n"
            "If mongo server <v4.4, ensure that C:\\Program Files\\MongoDB\\Server\\<ServerVersion>\\bin\\ \n"
            "has been added to the environment path.\n"
        )
        print("..................Upload failed..................")
    except subprocess.CalledProcessError as exc:
        print("Status : FAIL", exc.returncode, exc.output)
        raise exc


def load_mongo_col(col: Collection) -> dict:
    """Load the pymongo collection to a dictionary.

    In the dictionary. The key will be the '_id' and in each value which is a dictionary there will also be a
    key '_id' so that the structure will be the same as the filesystem collection.

    Parameters
    ----------
    col : Collection
        The mongodb collection.

    Returns
    -------
    dct : dict
        A dictionary with all the info in the collection.
    """
    return {doc["_id"]: doc for doc in col.find({})}


def doc_cleanup(doc: dict):
    doc = bson_cleanup(doc)
    doc["_id"].replace(".", "")
    return doc


def bson_cleanup(doc: dict):
    """
    This method should be used prior to updating or adding a document to a collection in mongo. Specifically, this
    replaces all periods in keys and _id value with a blank, and changes datetime.date to an iso string. It does so
    recursively for nested dictionaries.

    Parameters
    ----------
    doc

    Returns
    -------
    doc

    """

    def change_keys_id_and_date(obj, convert):
        """
        Recursively goes through the dictionary obj and replaces keys with the convert function.
        """
        if isinstance(obj, datetime.date):
            # Mongo cannot handle datetime.date format,
            # but we have infrastructure to handle iso date strings present
            # Do not convert to datetime.datetime. Mongo can handle this, but regolith cannot.
            return obj.isoformat()
        if isinstance(obj, (str, int, float)):
            return obj
        if isinstance(obj, dict):
            new = obj.__class__()
            for k, v in obj.items():
                new[convert(k)] = change_keys_id_and_date(v, convert)
        elif isinstance(obj, (list, set, tuple)):
            new = obj.__class__(change_keys_id_and_date(v, convert) for v in obj)
        else:
            return obj
        return new

    def convert(k):
        return k.replace(".", "-")

    doc = change_keys_id_and_date(doc, convert)
    return doc


@fallback(ON_PYMONGO_V2, None)
class InsertOneProxy(object):
    def __init__(self, inserted_id, acknowledged):
        self.inserted_id = inserted_id
        self.acknowledged = acknowledged


class MongoClient:
    """A client backed by MongoDB.

    The mongodb server will be automatically opened when the client is initiated.

    Attributes
    ----------
    rc : RunControl
        The RunControl. It may include the 'mongohost' attribute to initiate the client.

    client : MongoClient
        The mongo client. It is initiate from the 'mongohost' attribute if it exists in rc. Otherwise,
        it will be initiated from the 'localhost'.

    proc : Popen
        The Popen of 'mongod --dpath <mongodbpath>'. The 'mongodbpath' is from rc.
    """

    def __init__(self, rc):
        if not MONGO_AVAILABLE:
            raise RuntimeError("MongoDB is not available on the current system.")
        self.rc = rc
        self.client = None
        self.proc = None
        self.dbs = defaultdict(lambda: defaultdict(dict))
        self.chained_db = dict()
        self.closed = True
        self.local = True

    def _preclean(self):
        mongodbpath = self.rc.mongodbpath
        if os.path.isdir(mongodbpath):
            shutil.rmtree(mongodbpath)
        os.makedirs(mongodbpath)

    def _startserver(self):
        mongodbpath = self.rc.mongodbpath
        self.proc = subprocess.Popen(
            ["mongod", "--fork", "--syslog", "--dbpath", mongodbpath], universal_newlines=True
        )
        print("mongod pid: {0}".format(self.proc.pid), file=sys.stderr)

    def is_alive(self):
        """Returns whether or not the client is alive and available to
        send/receive data.
        """
        if self.client is None:
            return False
        elif ON_PYMONGO_V2:
            return self.client.alive()
        elif ON_PYMONGO_V3:
            alive = False
            if self.local is False:
                from pymongo.errors import ConnectionFailure

                try:
                    # The ismaster command is cheap and does not require auth.
                    self.client.admin.command("ismaster")
                    alive = True
                except ConnectionFailure:
                    print("Server not available")
                    alive = False
            else:
                cmd = ["mongostat", "--host", "localhost", "-n", "1"]
                try:
                    subprocess.check_call(cmd, stderr=subprocess.STDOUT)
                    alive = True
                except subprocess.CalledProcessError as exc:
                    print("Status : FAIL", exc.returncode, exc.output)
                    alive = False
            return alive
        else:
            return False

    def open(self):
        """Opens the database client"""
        if self.closed:
            rc = self.rc
            mongo_dbs_filter = filter(
                lambda db: db["backend"] == "mongo" or db["backend"] == "mongodb", rc.databases
            )
            mongo_dbs_list = list(mongo_dbs_filter)
            host = None
            if hasattr(rc, "host"):
                host = rc.host
            else:
                for db in mongo_dbs_list:
                    if host is not None:
                        if host != db["url"]:
                            print("WARNING: Multiple mongo clusters not supported. Use single cluster per rc.")
                            return
                    host = db["url"]
            if host is not None:
                if "+srv" in host:
                    self.local = False
            elif "dst_url" in rc.databases[0]:
                if "+srv" in rc.databases[0]["dst_url"]:
                    self.local = False
            # Currently configured such that password deemed unnecessary for strictly local mongo instance
            password_req = not self.local
            if password_req:
                try:
                    password = rc.mongo_db_password
                    if host is not None:
                        host = host.replace("pwd_from_config", urllib.parse.quote_plus(password))
                        host = host.replace("uname_from_config", urllib.parse.quote_plus(rc.mongo_id))
                    elif "dst_url" in rc.databases[0]:
                        rc.databases[0]["dst_url"] = rc.databases[0]["dst_url"].replace(
                            "pwd_from_config", urllib.parse.quote_plus(password)
                        )
                        rc.databases[0]["dst_url"] = rc.databases[0]["dst_url"].replace(
                            "uname_from_config", urllib.parse.quote_plus(rc.mongo_id)
                        )
                        host = rc.databases[0]["dst_url"]
                except AttributeError:
                    print(
                        "ERROR:\n"
                        "Add a username and password to user.json in "
                        "user/.config/regolith/user.json with the keys\n"
                        "mongo_id and mongo_db_password respectively.\n\n"
                        "'uname_from_config' and 'pwd_from_config' can/should stand in for these field in the\n"
                        "mongo URL string in regolithrc.json.\n"
                    )
            self.client = pymongo.MongoClient(host, authSource="admin")
            if not self.is_alive():
                if self.local:
                    # we need to wait for the server to startup
                    self._preclean()
                    self._startserver()
                    time.sleep(0.1)
                else:
                    raise ConnectionError(
                        "Mongo server exists, communication refused. Potential TLS issue.\n"
                        "Attempt the following in regolith env bash terminal:\n"
                        'export SSL_CERT_FILE=$(python -c "import certifi; print(certifi.where())")'
                    )
            self.closed = False

    def load_database(self, db: dict):
        """Load the database information from mongo database.

        It populate the 'dbs' attribute with a dictionary like {database: {collection: docs_dict}}.

        Parameters
        ----------
        db : dict
            The dictionary of data base information, such as 'name'.
        """
        dbs: dict = self.dbs
        client: pymongo.MongoClient = self.client
        from pymongo.errors import OperationFailure

        try:
            mongodb = client[db["name"]]
        except OperationFailure:
            print("WARNING: Database name provided in regolithrc.json not found in mongodb")
        try:
            for colname in [
                coll
                for coll in mongodb.list_collection_names()
                if coll not in db["blacklist"] and len(db["whitelist"]) == 0 or coll in db["whitelist"]
            ]:
                col = mongodb[colname]
                dbs[db["name"]][colname] = load_mongo_col(col)
        except OperationFailure as fail:
            print("Mongo's Error Message:" + str(fail) + "\n")
            print("The user does not have permission to access " + db["name"] + "\n\n")
            print(
                "If erroneous, the role/mongo-account utilized likely is only read/write. List collection \n"
                "permission as well as finding is needed"
            )
        return

    def import_database(self, db: dict):
        """Import the database from filesystem to the mongo backend.

        Parameters
        ----------
        db : dict
            The dictionary of data base information, such as 'name'.
        """
        host = getattr(self.rc, "host", None)
        uri = db.get("dst_url", None)
        # Catch the easy/common regolith rc error of putting the db uri as localhost rather than host
        if uri == "localhost":
            uri = None
            host = "localhost"
        dbpath = dbpathname(db, self.rc)
        dbname = db["name"]
        import_jsons(dbpath, dbname, host=host, uri=uri)
        import_yamls(dbpath, dbname, host=host, uri=uri)
        return

    def export_database(self, db: dict):
        """Exports the database from mongo backend to the filesystem.

        Parameters
        ----------
        db : dict
            The dictionary of data base information, such as 'name'.
        """
        host = getattr(self.rc, "host", None)
        uri = db.get("dst_url", None)
        # Catch the easy/common regolith rc error of putting the db uri as localhost rather than host
        if uri == "localhost":
            uri = None
            host = "localhost"
        dbpath = os.path.abspath(dbpathname(db, self.rc))
        dbname = db["name"]
        for collection in self.dbs[dbname].keys():
            export_json(collection, dbpath, dbname, host=host, uri=uri)
        return

    def dump_database(self, db):
        """Dumps a database dict via mongoexport."""
        dbpath = dbpathname(db, self.rc)
        os.makedirs(dbpath, exist_ok=True)
        to_add = []
        colls = self.client[db["name"]].collection_names(include_system_collections=False)
        for collection in colls:
            f = os.path.join(dbpath, collection + ".json")
            cmd = [
                "mongoexport",
                "--db",
                db["name"],
                "--collection",
                collection,
                "--out",
                f,
            ]
            try:
                subprocess.check_call(cmd, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as exc:
                print("Status : FAIL", exc.returncode, exc.output)
                raise exc
            to_add.append(os.path.join(db["path"], collection + ".json"))
        return to_add

    def close(self):
        """Closes the database connection."""
        self.closed = True
        return

    def keys(self):
        return self.client.list_database_names()

    def __getitem__(self, key):
        return self.client[key]

    def collection_names(self, dbname, include_system_collections=True):
        """Returns the collection names for the database name."""
        return self.client[dbname].collection_names()

    def all_documents(self, collname, copy=True):
        """Returns an iterable over all documents in a collection."""
        if copy:
            return deepcopy(self.chained_db.get(collname, {})).values()
        return self.chained_db.get(collname, {}).values()

    def insert_one(self, dbname, collname, doc):
        """Inserts one document to a database/collection."""
        doc = doc_cleanup(doc)
        valid, potential_error = validate_doc(collname, doc, self.rc)
        if not valid:
            raise ValueError(potential_error)
        coll = self.client[dbname][collname]
        if ON_PYMONGO_V2:
            i = coll.insert(doc)
            return InsertOneProxy(i, True)
        else:
            return coll.insert_one(doc)

    def insert_many(self, dbname, collname, docs):
        """Inserts many documents into a database/collection."""
        docs = [doc_cleanup(doc) for doc in docs]

        screened_docs = []
        full_error = ""
        for doc in docs:
            valid, potential_error = validate_doc(collname, doc, self.rc)
            if not valid:
                screened_docs.append(doc)
                full_error += potential_error
                full_error += "\n"
        if len(screened_docs) != 0:
            print("The following documents failed validation and were not uploaded\n")
            for doc in screened_docs:
                print(full_error)
                if doc in docs:
                    docs.remove(doc)
        coll = self.client[dbname][collname]
        if ON_PYMONGO_V2:
            return coll.insert(docs)
        else:
            return coll.insert_many(docs)

    def delete_one(self, dbname, collname, doc):
        """Removes a single document from a collection"""
        coll = self.client[dbname][collname]
        doc = doc_cleanup(doc)
        if ON_PYMONGO_V2:
            return coll.remove(doc, multi=False)
        else:
            return coll.delete_one(doc)

    def find_one(self, dbname, collname, filter):
        """Finds the first document matching filter."""
        filter["_id"].replace(".", "")
        coll = self.client[dbname][collname]
        doc = coll.find_one(filter)
        return doc

    def update_one(self, dbname, collname, filter, update, **kwargs):
        """Updates one document."""
        filter["_id"].replace(".", "")
        doc = self.find_one(dbname, collname, filter)
        newdoc = dict(filter if doc is None else doc)
        newdoc.update(update)
        valid, potential_error = validate_doc(collname, newdoc, self.rc)
        if not valid:
            raise ValueError(potential_error)
        coll = self.client[dbname][collname]
        update = bson_cleanup(update)
        if ON_PYMONGO_V2:
            doc = coll.find_one(filter)
            if doc is None:
                if not kwargs.get("upsert", False):
                    raise RuntimeError("could not update non-existing document")
                newdoc = dict(filter)
                newdoc.update(update["$set"])
                return self.insert_one(dbname, collname, newdoc)
            return coll.update(doc, update, **kwargs)
        else:
            return coll.find_one_and_update(filter, {"$set": update}, **kwargs)



"""Builder for Grade Reports."""

import os
import pdb
import sys
import traceback
from itertools import groupby

import numpy as np

try:
    import scipy.stats as st
except ImportError:
    st = None

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.tools import all_docs_from_collection


class GradeReportBuilder(LatexBuilderBase):
    btype = "grades"
    needed_colls = ["grades", "courses", "assignments"]

    def construct_global_ctx(self):
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["sum"] = sum
        gtx["zip"] = zip
        gtx["range"] = range
        gtx["id_key"] = lambda x: x["_id"]
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["grades"] = list(all_docs_from_collection(rc.client, "grades"))
        gtx["courses"] = list(all_docs_from_collection(rc.client, "courses"))
        gtx["assignments"] = list(all_docs_from_collection(rc.client, "assignments"))

    def render(self, tname, fname, **kwargs):
        template = self.env.get_template(tname)
        ctx = dict(self.gtx)
        ctx.update(kwargs)
        ctx["rc"] = ctx.get("rc", self.rc)
        ctx["static"] = ctx.get("static", os.path.relpath("static", os.path.dirname(fname)))
        ctx["root"] = ctx.get("root", os.path.relpath("/", os.path.dirname(fname)))
        try:
            result = template.render(ctx)
        except Exception:
            type, value, tb = sys.exc_info()
            traceback.print_exc()
            pdb.post_mortem(tb)
        with open(os.path.join(self.bldir, fname), "wt", encoding="utf-8") as f:
            f.write(result)

    def latex(self):
        for course in self.gtx["courses"]:
            if not course.get("active", True):
                continue
            course_id = course["_id"]
            stats = self.makestats(course)
            asgn = filter((lambda x: course_id in x["courses"]), self.gtx["assignments"])
            asgn = sorted(asgn, key=lambda x: x["category"])
            grouped_assignments = {
                k: sorted(i, key=lambda x: x["_id"]) for k, i in groupby(asgn, key=lambda x: x["category"])
            }

            student_wavgs = []
            students_kwargs = {}
            studs = []
            for student_id in course["students"]:
                studs.append(student_id)
                student_grade_list = list(
                    filter(
                        (lambda x: (x["student"] == student_id and x["course"] == course_id)),
                        self.gtx["grades"],
                    )
                )
                student_grades = {k: [] for k in grouped_assignments.keys()}
                for category, asngs in grouped_assignments.items():
                    for asgn in asngs:
                        for sg in student_grade_list:
                            if sg["assignment"] == asgn["_id"]:
                                student_grades[category].append(sg)
                                break
                        else:
                            student_grades[category].append(None)
                student_totals, student_wavg = self.maketotals(student_grades, grouped_assignments, course)
                student_wavgs.append(student_wavg)
                students_kwargs[student_id] = dict(
                    title=student_id,
                    stats=stats,
                    student_id=student_id,
                    course_id=course_id,
                    grouped_assignments=grouped_assignments,
                    student_grades=student_grades,
                    student_totals=student_totals,
                    student_wavg=student_wavg,
                )
            ordered_studs = sorted(studs, key=lambda x: (students_kwargs[x]["student_wavg"]), reverse=True)
            max_wavg = max(student_wavgs)
            curve = 1.0 - max_wavg
            # Make grades
            scale = course.get("scale", DEFAULT_LETTER_SCALE)
            for student_id in course["students"]:
                skw = students_kwargs[student_id]
                skw["student_letter_grade_raw"] = find_letter_grade(skw["student_wavg"], scale)
                skw["student_letter_grade_curved"] = find_letter_grade(skw["student_wavg"] + curve, scale)
            summary = [
                "{}: {:.2f}, Grade: {}".format(
                    stud,
                    students_kwargs[stud]["student_wavg"] * 100.0,
                    students_kwargs[stud]["student_letter_grade_raw"],
                )
                for stud in ordered_studs
            ]
            for stud in summary:
                print(stud)
            show_letter_plot = self.plot_letter_grades(students_kwargs, scale)
            # render PDF
            for student_id in course["students"]:
                base = self.basename(student_id, course_id)
                self.render(
                    "gradereport.tex",
                    base + ".tex",
                    p=student_id,
                    max_wavg=max_wavg,
                    curve=curve,
                    show_letter_plot=show_letter_plot,
                    **students_kwargs[student_id],
                )
                # TODO: this seems like something for the base class to handle
                self.pdf(base)

    def makestats(self, course):
        """Returns a dictionary of statistics for a course whose keys are
        the assignments and whose values are a (mean-problem, std-problem,
        mean-total, std-total) tuple.
        """
        scores = {}
        course_id = course["_id"]
        for grade in self.gtx["grades"]:
            if grade["course"] != course_id:
                continue
            assignment_id = grade["assignment"]
            if assignment_id not in scores:
                scores[assignment_id] = []
            scores[assignment_id].append(grade["scores"])
        stats = {}
        for assignment_id, data in scores.items():
            mu = np.mean(data, axis=0)
            sig = np.std(data, axis=0)
            max_score = np.max(data, axis=0)
            norm = st.norm(mu, sig)
            percent_above_60 = 1.0 - norm.cdf(0.6 * max_score)
            percent_above_80 = 1.0 - norm.cdf(0.8 * max_score)
            total = np.sum(data, axis=1)
            total_mu = np.mean(total, axis=0)
            total_sig = np.std(total, axis=0)
            total_max_score = np.max(total, axis=0)
            total_norm = st.norm(total_mu, total_sig)
            total_percent_above_60 = 1.0 - total_norm.cdf(0.6 * total_max_score)
            total_percent_above_80 = 1.0 - total_norm.cdf(0.8 * total_max_score)
            stats[assignment_id] = (
                mu,
                sig,
                max_score,
                percent_above_60,
                percent_above_80,
                total_mu,
                total_sig,
                total_max_score,
                total_percent_above_60,
                total_percent_above_80,
            )
        # handle stats for ungraded assignments
        for assignment in self.gtx["assignments"]:
            assignment_id = assignment["_id"]
            if assignment_id not in stats:
                n = len(assignment["points"])
                z = (0,) * n
                stats[assignment_id] = (z, z, z, z, z, 0, 0, 0, 0, 0)
        return stats

    @staticmethod
    def basename(student_id, course_id):
        """Returns the base file name for a student in a course."""
        name = student_id.replace(".", "").replace(" ", "")
        name += "-" + course_id
        return name

    def maketotals(self, student_grades, grouped_assignments, course):
        """Makes total grades."""
        totals = []
        totalweight = totalfrac = 0.0
        for category in student_grades.keys():
            cat = [category]
            sgtot = catmax = 0
            for sg, asgn in zip(student_grades[category], grouped_assignments[category]):
                if sg is not None:
                    sgtot += sum(sg["scores"])
                catmax += sum(asgn["points"])
            cat.append(sgtot)
            cat.append(catmax)
            weight = course["weights"][category]
            cat.append(weight)
            totalweight += weight
            catfrac = sgtot / catmax
            # cat.append(catfrac)
            wfrac = catfrac * weight
            cat.append(wfrac)
            totalfrac += wfrac
            totals.append(cat)
        wtotal = 0.0 if not totalweight else totalfrac / totalweight
        return sorted(totals), wtotal

    def plot_letter_grades(self, students_kwargs, scale):
        """Plots the letter grades in a histogram"""
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            return False
        bins = [x[1] for x in scale[::-1]]
        raws = []
        curveds = []
        for skw in students_kwargs.values():
            raws.append(skw["student_letter_grade_raw"])
            curveds.append(skw["student_letter_grade_curved"])
        rfreq = [raws.count(letter) for letter in bins]
        cfreq = [curveds.count(letter) for letter in bins]
        width = 1.0
        pos = np.arange(len(bins))
        f, (ax1, ax2) = plt.subplots(1, 2, sharey=True)
        ax1.set_xticks(pos + (width / 2))
        ax1.set_xticklabels(bins)
        ax1.set_xlabel("Raw Grade")
        ax1.set_ylabel("Number of Students")
        ax1.bar(pos, rfreq, width, color="yellow")
        ax1.grid(True)
        ax2.set_xticks(pos + (width / 2))
        ax2.set_xticklabels(bins)
        ax2.set_xlabel("Curved Grade")
        ax2.bar(pos, cfreq, width, color="green")
        ax2.grid(True)
        base = os.path.join(self.bldir, "student-letter-grade-dist")
        plt.savefig(base + ".png", bbox_inches="tight")
        plt.savefig(base + ".eps", bbox_inches="tight")
        return True


DEFAULT_LETTER_SCALE = (
    (0.97, "A+"),
    (0.93, "A"),
    (0.90, "A-"),
    (0.87, "B+"),
    (0.83, "B"),
    (0.80, "B-"),
    (0.77, "C+"),
    (0.73, "C"),
    (0.70, "C-"),
    (0.67, "D+"),
    (0.63, "D"),
    (0.60, "D-"),
    (-1.0, "F"),
)


def find_letter_grade(score, scale=DEFAULT_LETTER_SCALE):
    """Finds the letter grade from a score and a value."""
    for lower, letter in scale:
        if lower <= score:
            return letter
    return letter



"""Builder for websites."""

import string

from regolith.dates import date_to_float


def doc_date_key(document):
    """Convert a dict of Datetime object to float serialization of date info.

    Parameters
    ----------
    document:  dict
        the document that is expected to contain date-like information in "year" and/or "month"

    Returns
    -------
    the float serialization of the date information in the document

    """
    return date_to_float(document.get("year", 1970), document.get("month", "jan"))


def doc_date_key_high(document):
    """Convert a dict of highest Datetime object to float serialization of date info.

    Parameters
    ----------
    document: dict
        the document that is expected to contain date-like information in "end_year" and/or "end_month"

    Returns
    -------
    the float serialization of the end date information in the document

    """
    return date_to_float(document.get("year", 1970), document.get("month", "dec"))


def ene_date_key(document):
    """Convert a dict of ene Datetime object to float serialization of date info.

    Parameters
    ----------
    document: dict
        the document that is expected to contain date-like information in "end_year" and/or "end_month"

    Returns
    -------
    the float serialization of the date information in the document

    """
    return date_to_float(document.get("end_year", 4242), document.get("end_month", "dec"))


def category_val(document):
    """Convert the category of a document into string of category info.

    Parameters
    ----------
    document : dict
        The dict of all corresponding categories for objects

    Returns
    -------
    The string of the category item.

    """
    return document.get("category", "<uncategorized>")


def level_val(document):
    """Convert the level of a document into string of category info.

    Parameters
    ----------
    document: dict
        The document

    Returns
    -------
    The string representing the level.

    """
    return document.get("level", "<no-level>")


def id_key(document):
    """Convert the id-key of a document into a string

    Parameters
    ----------
    document: dict
        The document

    Returns
    -------
    The string of the _id

    """
    return document.get("_id", "")


def date_key(x):
    if "end_year" in x:
        v = date_to_float(x["end_year"], x.get("end_month", "jan"), x.get("end_day", 0))
    elif "year" in x:
        v = date_to_float(x["year"], x.get("month", "jan"), x.get("day", 0))
    elif "begin_year" in x:
        v = date_to_float(x["begin_year"], x.get("begin_month", "jan"), x.get("begin_day", 0))
    else:
        raise KeyError("could not find year in " + str(x))
    return v


POSITION_LEVELS = {
    "": -1,
    "editor": -1,
    "unknown": -1,
    "undergraduate research assistant": 1,
    "undergraduate researcher": 1,
    "intern": 1,
    "masters research assistant": 2.5,
    "masters researcher": 2.5,
    "visiting student": 2,
    "graduate research assistant": 3,
    "teaching assistant": 3,
    "research assistant": 2,
    "post-doctoral scholar": 4,
    "research fellow": 4,
    "assistant scientist": 4,
    "assistant lecturer": 4,
    "lecturer": 5,
    "research scientist": 4.5,
    "associate scientist": 6,
    "adjunct scientist": 5,
    "senior assistant lecturer": 5,
    "research associate": 5,
    "reader": 5,
    "ajunct professor": 5,
    "adjunct professor": 5,
    "consultant": 5,
    "programer": 5,
    "programmer": 5,
    "visiting scientist": 5,
    "research assistant professor": 5,
    "assistant professor": 8,
    "assistant physicist": 8,
    "associate professor": 9,
    "associate physicist": 9,
    "professor emeritus": 9,
    "visiting professor": 9,
    "manager": 10,
    "director": 10,
    "scientist": 10,
    "engineer": 10,
    "physicist": 10,
    "professor": 11,
    "president": 10,
    "distinguished professor": 12,
}


def position_key(x):
    """Sorts a people based on their position in the research group."""
    pos = x.get("position", "").casefold()
    first_letter_last = x.get("name", "zappa").rsplit(None, 1)[-1][0].upper()
    try:
        backward_position = 26 - string.ascii_uppercase.index(first_letter_last)
    except ValueError:
        raise ValueError(f"{x.get('_id')} has improperly composed name")
    return POSITION_LEVELS.get(pos, -1), backward_position



"""Helper for updating milestones to the projecta collection.
It can update the status, type, and due date of a projectum.
It can add a new milestone to the projecta collection.
"""

import datetime as dt
from copy import deepcopy

import dateutil.parser as date_parser
from gooey import GooeyParser

from regolith.dates import get_due_date
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.schemas import alloweds
from regolith.tools import all_docs_from_collection, fragment_retrieval, get_uuid

TARGET_COLL = "projecta"
MILESTONE_TYPES = alloweds.get("MILESTONE_TYPES")
PROJECTUM_STATI = alloweds.get("PROJECTUM_STATI")


def subparser(subpi):
    date_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"

    subpi.add_argument("-i", "--milestone_uuid", help="The uuid of a milestone. " "Takes a full or partial uuid.")
    subpi.add_argument(
        "-p",
        "--projectum_id",
        help="The id of the projectum. If you "
        "opt for this the program will assume "
        "you are adding a new milestone "
        "to the specified projectum.",
    )
    subpi.add_argument(
        "-u", "--due-date", help="New due date of the milestone. " "Required for a new milestone.", **date_kwargs
    )
    # Do not delete --database arg
    subpi.add_argument(
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    subpi.add_argument("-f", "--finish", action="store_true", help="Finish milestone. ")
    subpi.add_argument("-n", "--name", help="Name of the milestone. " "Required for a new milestone.")
    subpi.add_argument("-o", "--objective", help="Objective of the milestone. " "Required for a new milestone.")
    subpi.add_argument(
        "-s",
        "--status",
        help="Status of the milestone/deliverable: "
        f"{*PROJECTUM_STATI, }. "
        "Defaults to proposed for a new milestone.",
    )
    subpi.add_argument(
        "-t",
        "--type",
        help="Type of the milestone: " f"{*MILESTONE_TYPES, } " "Defaults to meeting for a new milestone.",
    )
    subpi.add_argument(
        "-a",
        "--audience",
        nargs="+",
        help="Audience of the milestone. " "Defaults to ['lead', 'pi', 'group_members'] for a new milestone.",
    )
    subpi.add_argument("--notes", nargs="+", help="Any notes you want to add to the milestone.")
    subpi.add_argument("--date", help="The date that will be used for testing.", **date_kwargs)
    return subpi


class MilestoneUpdaterHelper(DbHelperBase):
    """Helper for updating milestones to the projecta collection."""

    # btype must be the same as helper target in helper.py
    btype = "u_milestone"
    needed_colls = [f"{TARGET_COLL}"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        rc.coll = f"{TARGET_COLL}"
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        updated, zero, multiple = [], [], []
        if rc.date:
            now = date_parser.parse(rc.date).date()
        else:
            now = dt.date.today()
        new_mil = []
        pdoc = {}
        if rc.projectum_id and rc.milestone_uuid:
            raise RuntimeError(
                "Detected both a uuid fragment and projectum id.\n"
                "You may enter either a milestone uuid or a projectum id but not both.\n"
                "Enter a milestone uuid to update an existing milestone, "
                "or a projectum id to add a new milestone to that projectum.\n"
            )
        if not rc.projectum_id and not rc.milestone_uuid:
            raise RuntimeError(
                "No milestone uuid or projectum id was entered.\n"
                "Enter a milestone uuid to update an existing milestone, "
                "or a projectum id to add a new milestone to that projectum.\n"
            )
        if rc.projectum_id:
            rc.projectum_id = rc.projectum_id.strip()
            target_prum = rc.client.find_one(rc.database, rc.coll, {"_id": rc.projectum_id})
            if not target_prum:
                pra = fragment_retrieval(self.gtx["projecta"], ["_id"], rc.projectum_id)
                print("Projectum not found. Projecta with similar names: ")
                for i in range(len(pra)):
                    print(f"{pra[i].get('_id')}")
                print(
                    f"Please rerun the helper specifying the complete ID.\n"
                    f"If your prum id looks correct, check that this id is in the collection "
                    f"in the database {rc.database}.\n"
                    f"If this is not the case, rerun with --database set to "
                    f"the database where the item is located."
                )
                return
            milestones = deepcopy(target_prum.get("milestones"))
            all_milestones = []
            for item in milestones:
                item["identifier"] = "milestones"
            all_milestones.extend(milestones)
            for i in all_milestones:
                i["due_date"] = get_due_date(i)
            all_milestones.sort(key=lambda x: x["due_date"], reverse=False)
            index_list = list(range(2, (len(all_milestones) + 2)))
            mil = {}
            if not rc.due_date or not rc.name or not rc.objective:
                raise RuntimeError("name, objective, and due date are required for a new milestone")
            mil.update({"due_date": rc.due_date})
            mil["due_date"] = get_due_date(mil)
            mil.update({"objective": rc.objective, "name": rc.name, "uuid": get_uuid()})
            if rc.audience:
                mil.update({"audience": rc.audience})
            else:
                mil.update({"audience": ["lead", "pi", "group_members"]})
            if rc.status:
                mil.update({"status": rc.status})
            else:
                mil.update({"status": "proposed"})
            if rc.notes:
                mil.update({"notes": rc.notes})
            if rc.type:
                if rc.type in MILESTONE_TYPES:
                    mil.update({"type": rc.type})
                else:
                    raise ValueError(
                        "The type you have specified is not recognized. \n"
                        "Please rerun your command adding '--type' \n"
                        f"and giving a type from this list:\n{MILESTONE_TYPES}\n"
                    )
            else:
                mil.update({"type": "meeting"})
            mil.update({"identifier": "milestones"})
            new_mil.append(mil)
            pdoc = {"milestones": mil}
            new_all = deepcopy(all_milestones)
            for i, j in zip(index_list, new_all):
                if j["identifier"] == "milestones":
                    new_mil.append(j)
            for mile in new_mil:
                del mile["identifier"]
            new_mil.sort(key=lambda x: x["due_date"], reverse=False)
            pdoc.update({"milestones": new_mil})
            rc.client.update_one(rc.database, rc.coll, {"_id": rc.projectum_id}, pdoc)
            updated.append(f"{rc.projectum_id} has been updated in projecta")
        else:
            pdoc, upd_mil, all_miles, id = {}, [], [], []
            target_mil = fragment_retrieval(self.gtx["projecta"], ["milestones"], rc.milestone_uuid)
            target_del = fragment_retrieval(self.gtx["projecta"], ["_id"], rc.milestone_uuid)
            target_ko = fragment_retrieval(self.gtx["projecta"], ["_id"], rc.milestone_uuid[2:])
            if target_mil and not target_del and not target_ko:
                for prum in target_mil:
                    milestones = prum["milestones"]
                    for milestone in milestones:
                        if milestone.get("uuid")[0 : len(rc.milestone_uuid)] == rc.milestone_uuid:
                            upd_mil.append(milestone)
                        else:
                            all_miles.append(milestone)
                    if upd_mil:
                        pid = prum.get("_id")
                        id.append(pid)
            if target_del:
                for prum in target_del:
                    if prum.get("_id")[0 : len(rc.milestone_uuid)] == rc.milestone_uuid:
                        deliverable = prum["deliverable"]
                        upd_mil.append(deliverable)
                    if upd_mil:
                        pid = prum.get("_id")
                        id.append(pid)
            if target_ko and rc.milestone_uuid[:2] == "ko":
                for prum in target_ko:
                    if prum.get("_id")[0 : len(rc.milestone_uuid) - 2] == rc.milestone_uuid[2:]:
                        kickoff = prum["kickoff"]
                        upd_mil.append(kickoff)
                    if upd_mil:
                        pid = prum.get("_id")
                        id.append(pid)
            if len(upd_mil) == 0:
                zero.append(rc.milestone_uuid)
            elif len(upd_mil) == 1:
                for dict in upd_mil:
                    pdoc.update(dict)
                if not pdoc.get("type") and not rc.type and not target_del and not target_ko:
                    raise ValueError(
                        f"Milestone ({rc.milestone_uuid}) does not have a type set and this is required.\n"
                        "Specify '--type' and rerun the helper to update this milestone.\n"
                    )
                if rc.type:
                    if rc.type in MILESTONE_TYPES:
                        pdoc.update({"type": rc.type})
                    else:
                        raise ValueError(
                            "The type you have specified is not recognized. \n"
                            "Please rerun your command adding '--type' \n"
                            f"and giving a type from this list:\n{MILESTONE_TYPES}\n"
                        )
                for i in all_miles:
                    i["due_date"] = get_due_date(i)
                if rc.finish:
                    rc.status = "finished"
                    pdoc.update({"end_date": now})
                    if pdoc.get("notes"):
                        notes = pdoc.get("notes", [])
                        notes_with_closed_items = [note.replace("()", "(x)", 1) for note in notes]
                        pdoc["notes"] = notes_with_closed_items
                    if pdoc.get("name"):
                        updated.append(
                            f"The milestone '{pdoc.get('name')}' has been marked as finished in prum {id[0]}."
                        )
                    else:
                        name = "deliverable"
                        updated.append(f"The milestone '{name}' has been marked as finished in prum {id[0]}.")
                if rc.audience:
                    pdoc.update({"audience": rc.audience})
                if rc.due_date:
                    pdoc.update({"due_date": rc.due_date})
                if rc.name and pdoc.get("name"):
                    pdoc.update({"name": rc.name})
                elif rc.name and not pdoc.get("name"):
                    print(f"Ignoring 'name' assignment for deliverable uuid ({rc.milestone_uuid})")
                if rc.objective and pdoc.get("name"):
                    pdoc.update({"objective": rc.objective})
                elif rc.objective and not pdoc.get("name"):
                    print(f"Ignoring 'objective' assignment for deliverable uuid ({rc.milestone_uuid})")
                if rc.status:
                    pdoc.update({"status": rc.status})
                if rc.notes:
                    pdoc.update({"notes": rc.notes})
                doc = {}
                pdoc["due_date"] = get_due_date(pdoc)
                all_miles.append(pdoc)
                all_miles.sort(key=lambda x: x["due_date"], reverse=False)
                if target_mil and not target_del and not target_ko:
                    doc.update({"milestones": all_miles})
                if target_del:
                    doc.update({"deliverable": pdoc})
                if target_ko and rc.milestone_uuid[:2] == "ko":
                    doc.update({"kickoff": pdoc})
                rc.client.update_one(rc.database, rc.coll, {"_id": id[0]}, doc)
                if not rc.finish:
                    updated.append(
                        f"The milestone uuid {rc.milestone_uuid} in {id[0]} has been updated in projecta."
                    )
            else:
                multiple.append(rc.milestone_uuid)
        if updated:
            for msg in updated:
                print(msg)
        else:
            print("Failed to update projecta.")
        if zero:
            print(
                f"No ids were found that match your milestone_uuid entry ({zero[0]}).\n"
                "Make sure you have entered the correct uuid or uuid fragment and rerun the helper.\n"
            )
        if multiple:
            print(
                f"Multiple ids match your milestone_uuid entry ({multiple[0]}).\n"
                "Try entering more characters of the uuid and rerun the helper.\n"
            )
        return



"""Helper for listing a summary of finished prums and progress on open prums.

Projecta are small bite-sized project quanta that typically will result in
one manuscript.
"""

import datetime

import dateutil.parser as date_parser
from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.schemas import PROJECTUM_ACTIVE_STATI, PROJECTUM_FINISHED_STATI, alloweds
from regolith.tools import all_docs_from_collection, get_pi_id, key_value_pair_filter

TARGET_COLL = "projecta"
HELPER_TARGET = "l_progress"

PROJECTUM_STATI = alloweds.get("PROJECTUM_STATI")


def subparser(subpi):
    listbox_kwargs = {}
    if isinstance(subpi, GooeyParser):
        listbox_kwargs["widget"] = "Listbox"

    subpi.add_argument("lead", help="Generate report for this project lead")
    subpi.add_argument("-v", "--verbose", action="store_true", help="increase verbosity of output")
    subpi.add_argument(
        "-s",
        "--stati",
        nargs="+",
        choices=PROJECTUM_STATI,
        help=f"Filter projecta for these stati."
        f" Default is {*(PROJECTUM_ACTIVE_STATI+PROJECTUM_FINISHED_STATI), }",
        default=PROJECTUM_ACTIVE_STATI + PROJECTUM_FINISHED_STATI,
        **listbox_kwargs,
    )
    # The --filter and --keys flags should be in every lister
    subpi.add_argument("-f", "--filter", nargs="+", help="Search this collection by giving key element pairs")
    subpi.add_argument(
        "-k",
        "--keys",
        nargs="+",
        help="Specify what keys to return values from when running "
        "--filter. If no argument is given the default is just the id.",
    )
    subpi.add_argument("--date", help="date used in testing. Defaults to " "today's date.")
    return subpi


class ProgressReportHelper(SoutHelperBase):
    """Helper for listing upcoming (and past) projectum milestones.

    Projecta are small bite-sized project quanta that typically will result in
    one manuscript.
    """

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET
    needed_colls = [f"{TARGET_COLL}"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        colls = [
            sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in self.needed_colls
        ]
        for db, coll in zip(self.needed_colls, colls):
            gtx[db] = coll
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def print_projectum(self, selected_projecta):
        rc = self.rc
        if selected_projecta == []:
            return
        selected_projecta.sort(key=lambda prum: prum.get("begin_date"), reverse=True)
        for p in selected_projecta:
            if rc.verbose:
                print(f"{p.get('_id')}")
                if p.get("deliverable"):
                    print(
                        f"  status: {p.get('status')}, begin_date: {p.get('begin_date')}, "
                        f"due_date: {p.get('deliverable').get('due_date')}"
                    )
                if p.get("status") == "finished":
                    print(f"  finished: {p.get('end_date')}")
                print(f"  description: {p.get('description')}")
                print(f"  log_url: {p.get('log_url')}")
                print("  team:")
                grp_members = None
                if p.get("group_members"):
                    grp_members = ", ".join(p.get("group_members"))
                collaborators = None
                if p.get("collaborators"):
                    collaborators = ", ".join(p.get("collaborators"))
                print(f"    group_members: {grp_members}")
                print(f"    collaborators: {collaborators}")
                d = p.get("deliverable")
                print("  deliverable:")
                audience = None
                if d.get("audience"):
                    audience = ", ".join(d.get("audience"))
                print(f"    audience: {audience}")
                iter, title = 1, "scope:"
                for scopum in d.get("scope", ["no scope"]):
                    print(f"    {title} {str(iter)}. {scopum}")
                    iter += 1
                    title = "      "
                print(f"    platform: {d.get('platform')}")
                print("  milestones:")
                for m in p.get("milestones"):
                    print(f"    {m.get('due_date')}: {m.get('name')}")
                    print(f"      objective: {m.get('objective')}")
                    print(f"      status: {m.get('status')}")
            else:
                print(f"{p.get('_id')}")
                if p.get("deliverable"):
                    print(
                        f"  status: {p.get('status')}, begin_date: {p.get('begin_date')}, "
                        f"due_date: {p.get('deliverable').get('due_date')}"
                    )
                    print(f"  description: {p.get('description')}")
                if p.get("status") == "finished":
                    print(f"    finished: {p.get('end_date')}")
                elif p.get("status") in PROJECTUM_ACTIVE_STATI:
                    print(f"  log_url: {p.get('log_url')}")
                    if p.get("milestones"):
                        print("  milestones:")
                    for m in p.get("milestones"):
                        print(
                            f"    due: {m.get('due_date')}, {m.get('name')}, "
                            f"type: {m.get('type')}, status: {m.get('status')}"
                        )
                        print(f"    objective: {m.get('objective')}")

    def sout(self):
        rc = self.rc
        if rc.filter:
            collection = key_value_pair_filter(self.gtx["projecta"], rc.filter)
        else:
            collection = self.gtx["projecta"]
        if not rc.date:
            now = datetime.date.today()
        else:
            now = date_parser.parse(rc.date).date()

        # remove checklist prums from the report
        collection = [
            prum for prum in collection if "checklist" not in prum.get("deliverable", {}).get("scope", [])
        ]

        title = f"\nProgress report for {rc.lead}, generated {now.isoformat()}"
        print(title)
        projecta = [valid_prum for valid_prum in collection if valid_prum.get("lead") == rc.lead]

        finishedp, proposedp, startedp, otherp = [], [], [], []
        for prum in projecta:
            if prum.get("status") == "finished":
                finishedp.append(prum)
            elif prum.get("status") == "proposed":
                proposedp.append(prum)
            elif prum.get("status") == "started":
                startedp.append(prum)
            else:
                otherp.append(prum)
        print("*************************[Orphan Projecta]*************************")
        for prum in otherp:
            print(f"{prum.get('_id')}, status: {prum.get('status')}")
        print("*************************[Finished Projecta]*************************")
        for prum in finishedp:
            print(f"{prum.get('_id')}, grant: {prum.get('grants')}")
            print(f"  description: {prum.get('description')}")
            print(f"  finished: {prum.get('end_date')}")
        print("*************************[Proposed Projecta]*************************")
        self.print_projectum(proposedp)
        print("*************************[In Progress Projecta]*************************")
        self.print_projectum(startedp)



"""Builder for Current and Pending Reports."""

import datetime as dt
from copy import copy

from nameparser import HumanName

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.dates import get_dates, is_current
from regolith.fsclient import _id_key
from regolith.sorters import position_key
from regolith.tools import all_docs_from_collection, filter_grants, fuzzy_retrieval, merge_collections_all


def is_pending(status):
    return status in "pending"


def is_declined(status):
    return status in "declined"


class CPBuilder(LatexBuilderBase):
    """Build current and pending report from database entries"""

    btype = "current-pending"
    needed_colls = ["groups", "people", "grants", "proposals"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["people"] = list(
            sorted(
                all_docs_from_collection(rc.client, "people"),
                key=position_key,
                reverse=True,
            )
        )
        gtx["groups"] = list(sorted(all_docs_from_collection(rc.client, "groups"), key=_id_key))
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def latex(self):
        """Render latex template"""
        rc = self.rc
        for group in self.gtx["groups"]:
            self.gtx["grants"] = list(sorted(all_docs_from_collection(rc.client, "grants"), key=_id_key))
            self.gtx["proposals"] = list(sorted(all_docs_from_collection(rc.client, "proposals"), key=_id_key))
            grp = group["_id"]
            pi = fuzzy_retrieval(self.gtx["people"], ["_id", "aka", "name"], group["pi_name"])
            pinames = pi["name"].split()
            piinitialslist = [i[0] for i in pinames]
            pi["initials"] = "".join(piinitialslist).upper()

            grants = merge_collections_all(self.gtx["proposals"], self.gtx["grants"], "proposal_id")
            for g in grants:
                g["end_date"] = get_dates(g).get("end_date")
                g["begin_date"] = get_dates(g).get("begin_date", dt.date(1900, 1, 2))
                for person in g.get("team", []):
                    rperson = fuzzy_retrieval(self.gtx["people"], ["_id", "aka", "name"], person["name"])
                    if rperson:
                        person["name"] = rperson["name"]
                if g.get("budget"):
                    amounts = [i.get("amount") for i in g.get("budget")]
                    g["subaward_amount"] = sum(amounts)

            current_grants = [
                dict(g)
                for g in grants
                if is_current(g)
                and not is_pending(g.get("status", "None"))
                and not is_declined(g.get("status", "None"))
            ]
            current_grants, _, _ = filter_grants(current_grants, {pi["name"]}, pi=False, multi_pi=True)
            for g in current_grants:
                if g.get("budget"):
                    amounts = [i.get("amount") for i in g.get("budget")]
                    g["subaward_amount"] = sum(amounts)

            pending_grants = [g for g in grants if is_pending(g.get("status", "None"))]
            for g in pending_grants:
                for person in g["team"]:
                    rperson = fuzzy_retrieval(self.gtx["people"], ["aka", "name"], person["name"])
                    if rperson:
                        person["name"] = rperson["name"]
            pending_grants, _, _ = filter_grants(pending_grants, {pi["name"]}, pi=False, multi_pi=True)
            summed_grants = pending_grants + current_grants
            for grant in summed_grants:
                grant.update(
                    award_start_date="{}/{}/{}".format(
                        grant.get("begin_date").month,
                        grant.get("begin_date").day,
                        grant.get("begin_date").year,
                    ),
                    award_end_date="{}/{}/{}".format(
                        grant.get("end_date").month,
                        grant.get("end_date").day,
                        grant.get("end_date").year,
                    ),
                )
            badids = [i["_id"] for i in current_grants if not i.get("cpp_info", {}).get("cppflag", "")]
            # badids_pending = ([i["_id"] for i in pending_grants if
            #           not i.get('cpp_info',{}).get('cppflag', "")])
            if badids:
                print(f"these grants have a problem with ccp_info: {*badids, }")

            iter = copy(current_grants)
            for grant in iter:
                if grant["_id"] in badids:
                    current_grants.remove(grant)
            piname = HumanName(pi["name"])
            outfile = "current-pending-{}-{}".format(grp, piname.last.lower())

            self.render(
                "current_pending.tex",
                outfile + ".tex",
                pi=pi,
                pending=pending_grants,
                current=current_grants,
                pi_upper=pi["name"].upper(),
                group=group,
            )
            self.pdf(outfile)



"""Builder for Resumes."""

import os

import openpyxl

from regolith.builders.basebuilder import BuilderBase
from regolith.dates import get_dates
from regolith.sorters import position_key
from regolith.tools import all_docs_from_collection, fuzzy_retrieval, month_and_year


class ReimbursementBuilder(BuilderBase):
    """Build reimbursement from database entries"""

    btype = "reimb"
    needed_colls = ["expenses", "people", "grants"]

    def __init__(self, rc):
        super().__init__(rc)
        # TODO: templates for other universities?
        self.template = os.path.join(os.path.dirname(os.path.dirname(__file__)), "templates", "reimb.xlsx")
        self.cmds = ["excel"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        # openpyxl is soooo slow, so only allow it to be run when a person
        # (or people list) is specified
        if not rc.people:
            raise ValueError(
                "Missing person for the reimbursement.  Please "
                "rerun specifying --people and a person or list "
                "of people"
            )
        gtx["month_and_year"] = month_and_year
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        for n in ["expenses", "grants"]:
            gtx[n] = list(all_docs_from_collection(rc.client, n))
        gtx["all_docs_from_collection"] = all_docs_from_collection

    def excel(self):
        gtx = self.gtx
        rc = self.rc
        if isinstance(rc.people, str):
            rc.people = [rc.people]
        for ex in gtx["expenses"]:
            payee = fuzzy_retrieval(gtx["people"], ["name", "aka", "_id"], ex["payee"])
            chosen_ones = [fuzzy_retrieval(gtx["people"], ["name", "aka", "_id"], one) for one in rc.people]
            if ex["payee"] != "direct_billed":
                for chosen_one in chosen_ones:
                    if not payee:
                        print(f"WARNING: payee {ex['payee']} not found in " f"people coll")
                    elif payee.get("name") != chosen_one.get("name"):
                        continue
                    # open the template
                    if isinstance(ex["grants"], str):
                        ex["grants"] = [ex["grants"]]
                        grant_fractions = [1.0]
                    else:
                        grant_fractions = [float(percent) / 100.0 for percent in ex["grant_percentages"]]

                    wb = openpyxl.load_workbook(self.template)
                    ws = wb["T&B"]

                    grants = []
                    for grant_label in ex["grants"]:
                        grant = fuzzy_retrieval(gtx["grants"], ["alias", "name", "_id"], grant_label)
                        if not grant:
                            raise ValueError(f"no grant found with label {grant_label}")
                        else:
                            grants.append(grant)
                    ha = payee["home_address"]
                    ws["B17"] = payee["name"]
                    ws["B20"] = ha["street"]
                    ws["B23"] = ha["city"]
                    ws["G23"] = ha["state"]
                    ws["L23"] = ha["zip"]
                    ws["B36"] = ex["overall_purpose"]
                    j = 42
                    total_amount = 0
                    item_ws = wb["T&B"]
                    purpose_column = 4
                    ue_column = 13
                    se_column = 16
                    dates = []
                    for i, item in enumerate(ex["itemized_expenses"]):
                        r = j + i
                        expdates = get_dates(item)
                        if r > 49:
                            item_ws = wb["Extra_Page"]
                            j = 0
                            r = j + i
                            purpose_column = 5
                            ue_column = 12
                            se_column = 14
                        dates.append(expdates.get("date"))
                        item_ws.cell(row=r, column=2, value=i)
                        item_ws.cell(row=r, column=3, value=expdates.get("date").strftime("%x"))
                        item_ws.cell(row=r, column=purpose_column, value=item["purpose"])
                        item_ws.cell(
                            row=r,
                            column=ue_column,
                            value=item.get("unsegregated_expense", 0),
                        )
                        try:
                            total_amount += item.get("unsegregated_expense", 0)
                        except TypeError:
                            if item.get("unsegregated_expense", 0) == "tbd":
                                print("WARNING: unsegregated expense in {} is " "tbd".format(ex["_id"]))
                                item["unsegregated_expense"] = 0
                            else:
                                raise TypeError("unsegregated expense in {} is not " "a number".format(ex["_id"]))

                        item_ws.cell(row=r, column=se_column, value=item.get("segregated_expense", 0))

                    i = 0
                    if abs(sum([fraction * total_amount for fraction in grant_fractions]) - total_amount) >= 0.01:
                        raise RuntimeError("grant percentages do not sum to 100")
                    for grant, fraction in zip(grants, grant_fractions):
                        nr = grant.get("account", "")
                        row = 55 + i
                        location = "C{}".format(row)
                        location2 = "K{}".format(row)
                        ws[location] = nr
                        ws[location2] = total_amount * float(fraction)
                        i += 1

                    if ex.get("expense_type", "business") == "business":
                        spots = ("G10", "L11", "O11")
                    else:
                        spots = ("G7", "L8", "O8")

                    ws[spots[0]] = "X"
                    ws[spots[1]] = min(dates).strftime("%x")
                    ws[spots[2]] = max(dates).strftime("%x")

                    wb.save(os.path.join(self.bldir, ex["_id"] + ".xlsx"))



import datetime as dt

from regolith.dates import get_dates
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.tools import all_docs_from_collection

TARGET_COLL = "expenses"


def subparser(subpi):
    subpi.add_argument("payee", help="payee id for the expense")
    return subpi


class ReimbstatusHelper(SoutHelperBase):
    """Helper for reimbstatus"""

    # btype must be the same as helper target in helper.py
    btype = "reimbstatus"
    needed_colls = [f"{TARGET_COLL}"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["expenses"] = sorted(all_docs_from_collection(rc.client, "expenses"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc
        exps = self.gtx["expenses"]
        reimb, sub, unsub, future, unknown = [], [], [], [], []
        for expense in exps:
            if expense.get("payee") == rc.payee:
                dates = get_dates(expense)
                expense["end_date"] = dates["end_date"]
                expense["begin_date"] = dates["begin_date"]
                expense["begin_month"] = dates["begin_date"].month
                expense["end_month"] = dates["end_date"].month
                for j in expense.get("itemized_expenses"):
                    if j.get("unsegregated_expense") == "tbd":
                        print("WARNING: An expense in {} is tbd".format(expense["_id"]))
                        j["unsegregated_expense"] = 0
                    if j.get("exchange_rate"):
                        try:
                            j["unsegregated_expense"] = j.get("unsegregated_expense") / j.get("exchange_rate")
                        except TypeError:
                            print(
                                "exchange rate correction failed for {}, with "
                                "expense: {} rate: {}".format(
                                    expense["_id"], j.get("unsegregated_expense"), j.get("exchange_rate")
                                )
                            )
                        j["segregated_expense"] = j.get("segregated_expense") / j.get("exchange_rate")
                        j["prepaid_expense"] = j.get("prepaid_expense") / j.get("exchange_rate")
                if expense.get("status") == "reimbursed":
                    reimb.append(expense)
                elif expense.get("status") == "submitted":
                    sub.append(expense)
                elif expense.get("status") == "unsubmitted":
                    if expense["end_date"] < dt.datetime.today().date():
                        unsub.append(expense)
                    else:
                        future.append(expense)
                elif expense.get("status") == "cancelled":
                    pass
                else:
                    unknown.append(expense.get("_id"))
        sorted_reimb = sorted(reimb, key=lambda i: i["end_date"])
        sorted_sub = sorted(sub, key=lambda i: i["end_date"])
        sorted_unsub = sorted(unsub, key=lambda i: i["end_date"])
        sorted_future = sorted(future, key=lambda i: i["end_date"])
        print("Reimbursed expenses:")
        for i in sorted_reimb:
            unseg = 0
            for j in i.get("itemized_expenses"):
                unseg = unseg + j.get("unsegregated_expense")
            for j in i.get("reimbursements", []):
                reimb_dates = get_dates(j)
                print(
                    " - {} - {} {} to {}"
                    ",".format(
                        i.get("end_date").isoformat().replace("-", "")[2:],
                        i.get("overall_purpose")[:59],
                        i.get("begin_date"),
                        i.get("end_date"),
                    )
                )
                grantstring = ", ".join(i.get("grants"))
                print(
                    f"   Requested: {unseg}, "
                    f"Reimbursed: {j.get('amount')}, Date: "
                    f"{reimb_dates.get('date', dt.date(1900, 1, 1).isoformat())}, "
                    f"Grants: {grantstring}"
                )
        print("\nSubmitted expenses:")
        for i in sorted_sub:
            unseg, seg = 0, 0
            for j in i.get("itemized_expenses"):
                unseg = unseg + j.get("unsegregated_expense")
                seg = seg + j.get("segregated_expense")
            total = seg + unseg
            for j in i.get("reimbursements", []):
                print(
                    " - {} - {} {} to {}"
                    ",".format(
                        i.get("end_date").isoformat().replace("-", "")[2:],
                        i.get("overall_purpose")[:59],
                        i.get("begin_date"),
                        i.get("end_date"),
                    )
                )
                if j.get("submission_date"):
                    when = j.get("submission_date")
                else:
                    when = "-".join(
                        (
                            str(j.get("submission_year")),
                            str(j.get("submission_month")),
                            str(j.get("submission_day")),
                        )
                    )
                print(
                    "   Expenses: unseg={:.2f}, Seg={:.2f}, Total={:.2f}, Where: {}, When: {}".format(
                        unseg, seg, total, j.get("where"), when
                    )
                )
            grantstring = ", ".join(i.get("grants"))
            print("   Grants: {}".format(grantstring))
            if isinstance(i.get("notes"), str):
                print(i.get("notes"))
            else:
                for note in i.get("notes"):
                    print("    - {}".format(note[:59]))
                    if len(note) > 60:
                        print("      {}".format(note[60:]))

        print("\nUnsubmitted expenses:")
        for i in sorted_unsub:
            unseg, seg = 0, 0
            for j in i.get("itemized_expenses"):
                unseg = unseg + j.get("unsegregated_expense")
                seg = seg + j.get("segregated_expense")
            total = seg + unseg
            for j in i.get("reimbursements", []):
                print(
                    " - {} - {} {} to {}"
                    ",".format(
                        i.get("end_date").isoformat().replace("-", "")[2:],
                        i.get("overall_purpose")[:59],
                        i.get("begin_date"),
                        i.get("end_date"),
                    )
                )
                print(
                    "   Expenses: unseg={:.2f}, Seg={:.2f}, Total={:.2f}, "
                    "Where: {}".format(
                        unseg,
                        seg,
                        total,
                        j.get("where"),
                    )
                )
            grantstring = ", ".join(i.get("grants"))
            print("   Grants: {}".format(grantstring))
        print("\nFuture expenses:")
        for i in sorted_future:
            unseg, seg = 0, 0
            for j in i.get("itemized_expenses"):
                unseg = unseg + j.get("unsegregated_expense")
                seg = seg + j.get("segregated_expense")
            total = seg + unseg
            for j in i.get("reimbursements", []):
                print(
                    " - {} - {} {} to {}"
                    ",".format(
                        i.get("end_date").isoformat().replace("-", "")[2:],
                        i.get("overall_purpose")[:59],
                        i.get("begin_date"),
                        i.get("end_date"),
                    )
                )
                print(
                    "   Expenses: unseg={:.2f}, Seg={:.2f}, Total={:.2f}, ".format(
                        unseg,
                        seg,
                        total,
                    )
                )

        if len(unknown) > 0:
            print("\nThese expenses have invalid statuses:")
            print(*unknown)



"""Builder for Current and Pending Reports."""

import datetime as dt

import dateutil.parser as date_parser
from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import all_docs_from_collection, get_tags

ALLOWED_TYPES = ["nsf", "doe", "other"]
ALLOWED_STATI = ["invited", "accepted", "declined", "downloaded", "inprogress", "submitted", "cancelled"]


def subparser(subpi):
    date_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"

    subpi.add_argument(
        "tags",
        help="list of tags, separated by spaces, to use "
        "to find papers in citations collection that "
        "will be added to the list. type 'all' if building"
        "lists from all tags in the db. OR logic is used "
        "so this will return all papers that "
        "contain this OR that tag in the tags string in "
        "citations.  for example, 'neutron superconductor' "
        "(without the quotes) will return all neutron "
        "and all superconductor papers.",
        nargs="+",
    )
    subpi.add_argument(
        "-t",
        "--title",
        help="A title for the list that will be "
        "rendered when the list is built. Required if "
        "this is new list.",
    )
    subpi.add_argument(
        "-p",
        "--purpose",
        help="The purpose or intended use for the reading. This will " "not be rendered when the list is built",
    )
    subpi.add_argument(
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    subpi.add_argument("--date", help="The date that will be used for testing.", **date_kwargs)
    return subpi


class GrpPubReadListAdderHelper(DbHelperBase):
    """Build a helper"""

    btype = "a_grppub_readlist"
    needed_colls = ["citations", "reading_lists"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        rc.coll = "reading_lists"
        gtx["citations"] = sorted(all_docs_from_collection(rc.client, "citations"), key=_id_key)
        gtx["reading_lists"] = sorted(all_docs_from_collection(rc.client, "reading_lists"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def build_reading_list_from_citation_tag(self, tag, update_date):
        rc = self.rc
        key = f"{tag.strip()}"
        rlist_doc = rc.client.find_one(rc.database, rc.coll, {"_id": key})
        if not rlist_doc:
            rlist_doc = {"_id": key, "date": update_date}
        # rebuild list from scratch always
        rlist_doc.update({"papers": []})
        if rc.purpose and rc.tags[0] != "all":
            rlist_doc.update({"purpose": rc.purpose})
        if rc.title and rc.tags[0] != "all":
            rlist_doc.update({"title": rc.title})
        elif not rlist_doc.get("title"):
            rlist_doc.update({"title": f"List built for tag {key} from citations collection"})
        for cite in self.gtx["citations"]:
            if tag in cite.get("tags", ""):
                new_paper = {
                    "text": cite.get("synopsis", f"no synopsis in citation " f"{cite.get('_id')}"),
                    "year": int(cite.get("year", 0)),
                }
                if cite.get("doi") is not None and cite.get("doi") != "tbd":
                    new_paper.update({"doi": cite.get("doi")})
                elif cite.get("url") is not None and cite.get("url") != "tbd":
                    new_paper.update({"url": cite.get("url")})
                else:
                    new_paper.update({"doi": "tbd"})
                rlist_doc["papers"].append(new_paper)
        # sort by year then remove year
        rlist_doc["papers"].sort(key=lambda pper: pper.get("year"))
        [ppr.pop("year") for ppr in rlist_doc["papers"]]
        # remove duplicates
        rlist_doc["papers"] = [dict(t) for t in {tuple(paper.items()) for paper in rlist_doc["papers"]}]
        return rlist_doc

    def db_updater(self):
        rc = self.rc
        if rc.tags[0].strip().lower() == "all":
            all_tags = get_tags(self.gtx["citations"])
        else:
            all_tags = rc.tags
        print(f"Making lists for tags:\n{all_tags}")
        if rc.date:
            update_date = date_parser.parse(rc.date).date()
        else:
            update_date = dt.date.today()

        for tag in all_tags:
            pdoc = self.build_reading_list_from_citation_tag(tag, update_date)
            rc.client.insert_one(rc.database, rc.coll, pdoc)
            print(f"{tag} has been added/updated in reading_lists")

        return



import copy
import os

from regolith.database import connect
from regolith.runcontrol import DEFAULT_RC, connect_db, filter_databases, load_rcfile


def test_connect_db(make_db):
    repo = make_db
    os.chdir(repo)
    rc = copy.copy(DEFAULT_RC)
    rc._update(load_rcfile("regolithrc.json"))
    filter_databases(rc)
    with connect(rc) as rc.client:
        expected_dbs = rc.client.dbs
        expected_chdb = rc.client.chained_db
    chained_db, dbs = connect_db(rc)
    assert chained_db == expected_chdb
    assert dbs == expected_dbs



import copy
import os
from collections.abc import Sequence
from pathlib import Path
from pprint import pprint

import pytest

from regolith.schemas import EXEMPLARS, SCHEMAS, validate


@pytest.mark.parametrize("key", SCHEMAS.keys())
def test_validation(key):
    if isinstance(EXEMPLARS[key], Sequence):
        for e in EXEMPLARS[key]:
            validate(key, e, SCHEMAS)
    else:
        validate(key, EXEMPLARS[key], SCHEMAS)


@pytest.mark.parametrize("key", SCHEMAS.keys())
def test_exemplars(key):
    if isinstance(EXEMPLARS[key], Sequence):
        for e in EXEMPLARS[key]:
            v = validate(key, e, SCHEMAS)
            assert v[0]
    else:
        v = validate(key, EXEMPLARS[key], SCHEMAS)
        if not v[0]:
            for vv, reason in v[1].items():
                print(vv, reason)
                print(type(EXEMPLARS[key][vv]))
                pprint(EXEMPLARS[key][vv])
        assert v[0]


BAD_PROJECTUM = {
    "_id": "sb_firstprojectum",
    "begin_date": "2020-04-28",
    "collaborators": ["aeinstein", "pdirac"],
    "deliverable": {
        "audience": ["beginning grad in chemistry"],
        "due_date": "2021-05-05",
        "success_def": "audience is happy",
        "scope": [
            "UCs that are supported or some other scope description " "if it is software",
            "sketch of science story if it is paper",
        ],
        "platform": "description of how and where the audience will access "
        "the deliverable.  Journal if it is a paper",
        "roll_out": [
            "steps that the audience will take to access and interact with " "the deliverable",
            "not needed for paper submissions",
        ],
        "notes": ["deliverable note"],
        "status": "proposed",
    },
}


def test_mongo_invalid_insertion(make_mongodb):
    # proof that valid insertion is allowed is provided by helper tests on mongo
    if make_mongodb is False:
        pytest.skip("Mongoclient failed to start")
    else:
        repo = Path(make_mongodb)
    from regolith.database import connect
    from regolith.runcontrol import DEFAULT_RC, load_rcfile

    os.chdir(repo)
    rc = copy.copy(DEFAULT_RC)
    rc.schemas = SCHEMAS
    rc._update(load_rcfile("regolithrc.json"))
    with connect(rc) as rc.client:
        only_database_in_test = rc.databases[0]["name"]
        try:
            rc.client.insert_one(only_database_in_test, "projecta", BAD_PROJECTUM)
        except ValueError as e:
            result = e.args[0]
    expected = (
        "ERROR in sb_firstprojectum:\n{'lead': ['required field'], 'status': ['required field']}"
        "\nNone\nNone\n---------------\n"
    )
    assert result == expected



"""Validator for meetings"""

import datetime as dt

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.tools import all_docs_from_collection, get_pi_id, validate_meeting

TARGET_COLL = "meetings"
HELPER_TARGET = "v_meetings"


def subparser(subpi):
    subpi.add_argument("-t", "--test", action="store_true", help="Testing flag for meeting validator")
    return subpi


class MeetingsValidatorHelper(SoutHelperBase):
    """Helper for validating the entries of the meetings.yml file"""

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET
    needed_colls = [f"{TARGET_COLL}", "institutions"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        try:
            if not rc.database:
                rc.database = rc.databases[0]["name"]
        except BaseException:
            pass
        colls = [
            sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in self.needed_colls
        ]
        for db, coll in zip(self.needed_colls, colls):
            gtx[db] = coll
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc
        if rc.test:
            print("Meeting validator helper")
            return
        date = dt.date.today()
        collection = self.gtx["meetings"]
        for meeting in collection:
            validate_meeting(meeting, date)
        return
